# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import itertools
from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class ORCIDiD:
    """An ORCID iD.

    Ensures that the id is valid during initialization.
    See https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier
    This class can be used with Pydantic models.

    Examples
    --------

        >>> from scippneutron.metadata import ORCIDiD
        >>> orcid_id = ORCIDiD('0000-0000-0000-0001')
        >>> str(orcid_id)
        'https://orcid.org/0000-0000-0000-0001'

    Or equivalently with an explicit resolver:

        >>> orcid_id = ORCIDiD('https://orcid.org/0000-0000-0000-0001')
        >>> str(orcid_id)
        'https://orcid.org/0000-0000-0000-0001'
    """

    __slots__ = ('_orcid_id',)

    def __init__(self, orcid_id: str | ORCIDiD) -> None:
        if isinstance(orcid_id, ORCIDiD):
            self._orcid_id: str = orcid_id._orcid_id
        else:
            self._orcid_id = _parse_id(orcid_id)

    def __str__(self) -> str:
        return self._orcid_id

    def __repr__(self) -> str:
        return f'ORCIDiD({self!s})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ORCIDiD):
            return self._orcid_id == other._orcid_id
        if isinstance(other, str):
            try:
                return self._orcid_id == _parse_id(other)
            except ValueError:  # other is not a valid ORCID iD
                return False
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        if (b := self.__eq__(other)) is NotImplemented:
            return NotImplemented
        return not b

    def __hash__(self) -> int:
        return hash(self._orcid_id)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            _parse_pydantic,
            core_schema.union_schema(
                [core_schema.is_instance_schema(ORCIDiD), core_schema.str_schema()]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.__str__, info_arg=False, return_schema=core_schema.str_schema()
            ),
        )


def _parse_pydantic(value: str | ORCIDiD) -> ORCIDiD:
    return ORCIDiD(value)


_ORCID_RESOLVER: str = 'https://orcid.org'


def _parse_id(value: str) -> str:
    parts = value.rsplit('/', 1)
    if len(parts) == 2:
        resolver, orcid_id = parts
        if resolver != _ORCID_RESOLVER:
            # Must be the correct ORCID URL.
            raise ValueError(
                f"Invalid ORCID URL: '{resolver}'. Must be '{_ORCID_RESOLVER}'"
            )
    else:
        value = f'{_ORCID_RESOLVER}/{value}'
        (orcid_id,) = parts
    _check_id(orcid_id)
    return value


def _check_id(orcid_id: str) -> None:
    segments = orcid_id.split('-')
    if len(segments) != 4 or not all(len(s) == 4 for s in segments):
        # Must have 4 blocks of 4 digits each.
        raise ValueError(f"Invalid ORCID iD: '{orcid_id}'. Incorrect structure.")
    if _orcid_id_checksum(segments) != orcid_id[-1]:
        # Checksum must match the last digit.
        raise ValueError(f"Invalid ORCID iD: '{orcid_id}'. Checksum does not match.")


def _orcid_id_checksum(segments: list[str]) -> str:
    total = 0
    for d in map(int, itertools.islice(itertools.chain(*segments), 4 * 4 - 1)):
        total = (total + d) * 2
    result = (12 - total % 11) % 11
    return 'X' if result == 10 else str(result)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc
import scippnexus as snx
from dateutil.parser import parse as parse_datetime

import scippneutron as scn
import scippneutron.data
from scippneutron import metadata


def test_measurement_from_nexus_entry() -> None:
    with snx.File(scn.data.get_path('PG3_4844_event.nxs')) as f:
        experiment = metadata.Measurement.from_nexus_entry(f['entry'])
    assert experiment.title == 'diamond cw0.533 4.22e12 60Hz [10x30]'
    assert experiment.run_number == '4844'
    assert experiment.experiment_id == 'IPTS-2767'
    assert experiment.start_time == parse_datetime('2011-08-12T11:50:17-04:00')
    assert experiment.end_time == parse_datetime('2011-08-12T13:22:05-04:00')
    assert experiment.experiment_doi is None


def test_measurement_from_individual_variables() -> None:
    experiment = metadata.Measurement(
        title=sc.scalar('The title'),
        run_number='12b',
        experiment_id=sc.scalar('EXP-1', unit=''),
    )
    assert experiment.title == 'The title'
    assert experiment.run_number == '12b'
    assert experiment.experiment_id == 'EXP-1'
    assert experiment.experiment_doi is None


def test_beamline_from_nexus_entry() -> None:
    with snx.File(scn.data.get_path('PG3_4844_event.nxs')) as f:
        beamline = metadata.Beamline.from_nexus_entry(f['entry'])
    assert beamline.name == 'POWGEN'
    assert beamline.facility is None
    assert beamline.site is None
    assert beamline.revision is None


def test_software_from_from_package_metadata_first_party() -> None:
    software = metadata.Software.from_package_metadata('scippneutron')
    expected = metadata.Software(
        name='scippneutron',
        version=scn.__version__,
        url='https://github.com/scipp/scippneutron',
        doi=None,  # Cannot be deduced
    )
    assert software == expected


def test_software_from_from_package_metadata_third_party() -> None:
    software = metadata.Software.from_package_metadata('scipp')
    assert software.name == 'scipp'
    assert software.version == sc.__version__
    assert software.url in (
        'https://github.com/scipp/scipp',  # properly deduced
        None,  # fallback for our conda packages
    )
    assert software.doi is None


def test_software_from_from_package_metadata_fails_when_package_not_installed() -> None:
    with pytest.raises(ModuleNotFoundError):
        metadata.Software.from_package_metadata('not-a-package')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import enum
import sys
from typing import Literal


class Byteorder(enum.Enum):
    little = "little"
    big = "big"

    @classmethod
    def parse(cls, value: Byteorder | Literal["native", "little", "big"]) -> Byteorder:
        if isinstance(value, Byteorder):
            return value
        if isinstance(value, str):
            if value == "native":
                return cls.native()
            return cls(value)
        raise ValueError(f"Invalid Byteorder: {value}")

    @classmethod
    def native(cls) -> Byteorder:
        return cls(sys.byteorder)

    def get(self) -> Literal["little", "big"]:
        match self:
            case Byteorder.little:
                return "little"
            case Byteorder.big:
                return "big"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import enum
from datetime import datetime
from typing import Annotated, Any

import scipp as sc
import scippnexus as snx
from dateutil.parser import parse as parse_datetime
from pydantic import BaseModel, BeforeValidator, EmailStr

from ._orcid import ORCIDiD


def _unpack_variable(value: object) -> Any:
    """Before validator to support passing scalar scipp variables as inputs."""
    if isinstance(value, sc.Variable):
        if value.dims:
            raise ValueError("Must be a scalar")
        if value.unit and value.unit != '':
            raise ValueError("Must be dimensionless or have no unit at all")
        return value.value
    return value


class Beamline(BaseModel):
    """A beamline / instrument.

    ``name`` should be the canonical spelling of the beamline name.
    The location of the beamline is split into ``facility`` and ``site``, where
    a 'facility' is located at a 'site'. For example:

    >>> from scippneutron.metadata import Beamline
    >>> beamline = Beamline(
    ...     name='Amor',
    ...     facility='SINQ',
    ...     site='PSI',
    ... )

    If there is no separate facility and site, either omit ``site`` or use the same
    value for both ``facility`` and ``site``. For example:

    >>> beamline = Beamline(
    ...     name='ESTIA',
    ...     facility='ESS',
    ...     site='ESS',  # can be omitted
    ... )

    If the beamline has been upgraded, provide a revision to indicate
    which version of the beamline was used.
    """

    name: Annotated[str, BeforeValidator(_unpack_variable)]
    """Name of the beamline."""
    facility: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Facility where the beamline is located."""
    site: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Site where the facility is located."""
    revision: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Revision of the beamline in case of upgrades."""

    @classmethod
    def from_nexus_entry(
        cls, entry: snx.Group, *, instrument_name: str | None = None
    ) -> Beamline:
        """Construct a Beamline object from a Nexus entry.

        NeXus does not have a standard method for specifying the facility, site, or
        revision. This function only sets those fields for known instruments.

        Parameters
        ----------
        entry:
            ScippNexus group for a NeXus entry.
            The entry needs to contain an ``NXinstrument`` with a 'name' field
            to identify the instrument.
        instrument_name:
            If the entry contains more than one ``NXinstrument`` group, this parameter
            must be the name of one of these groups.

        Returns
        -------
        :
            A Beamline object constructed from the given Nexus entry.
        """
        instrument = _get_unique_nexus_child(entry, snx.NXinstrument, instrument_name)
        instrument_name = _read_optional_nexus_string(instrument, 'name')
        if instrument_name is None:
            raise ValueError("No instrument name found in Nexus entry")

        facility, site = _guess_facility_and_site(instrument_name)

        return cls(
            name=instrument_name,
            facility=facility,
            site=site,
        )


class Measurement(BaseModel):
    """A single measurement.

    Terminology:

        - An "experiment" is the collection of all related measurements,
          typically done during one beamtime.
        - A "measurement" is a single step of data collection.
          It typically corresponds to a single NeXus file or a single entry
          in a NeXus file.

    The ``Measurement`` class represents a single measurement but also includes some
    information about the experiment that this measurement is part of.
    In particular, ``experiment_id`` and ``experiment_doi`` encode
    information about the experiment.
    *All* other fields encode information about a measurement; this includes
    ``start_time`` and ``end_time``.
    """

    title: Annotated[str | None, BeforeValidator(_unpack_variable)]
    """The title of the measurement."""
    run_number: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Run number of the measurement."""
    experiment_id: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """An ID for the experiment that this measurement is part of, e.g., proposal ID."""
    experiment_doi: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """A DOI for the experiment that this measurement is part of."""
    # These do not support conversion from variables because variables are naive w.r.t.
    # timezones and converting them to datetime.datetime would require extra info.
    start_time: datetime | None = None
    """Date and time when the measurement started."""
    end_time: datetime | None = None
    """Date and time when the measurement ended."""

    @classmethod
    def from_nexus_entry(cls, entry: snx.Group) -> Measurement:
        """Construct a Measurement object from a Nexus entry.

        Parameters
        ----------
        entry:
            ScippNexus group for a NeXus entry.

        Returns
        -------
        :
            An Measurement object constructed from the given Nexus entry.
        """
        return cls(
            title=_read_optional_nexus_string(entry, 'title'),
            run_number=_read_optional_nexus_string(entry, 'entry_identifier'),
            experiment_id=_read_optional_nexus_string(entry, 'experiment_identifier'),
            start_time=_read_optional_nexus_datetime(entry, 'start_time'),
            end_time=_read_optional_nexus_datetime(entry, 'end_time'),
            experiment_doi=None,
        )

    @property
    def run_number_maybe_int(self) -> int | str | None:
        """Return the run number as an int if possible."""
        try:
            return int(self.run_number)
        except ValueError:
            return self.run_number


class Person(BaseModel):
    """A person.

    .. attention::

        Please make sure that you don't specify personal details without permission!

    Specify an ORCID iD whenever possible to make sure the
    person can be uniquely identified.
    The name is always required but is often not unique.
    Other contact information like address or email address are less important
    and can usually be omitted when an ORCID iD is provided.
    """

    name: Annotated[str, BeforeValidator(_unpack_variable)]
    """Free form name of the person."""
    orcid_id: Annotated[ORCIDiD | None, BeforeValidator(_unpack_variable)] = None
    """ORCID iD of the person."""

    corresponding: bool = False
    """Whether the person is the corresponding / contact author."""
    owner: bool = True
    """Whether the person owns the data."""
    role: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """The role that the person played in collecting / processing the data.

    `NeXus <https://manual.nexusformat.org/classes/base_classes/NXuser.html#nxuser>`_
    and
    `CIF <https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic#L15167>`_
    list possible roles.
    """

    address: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Physical (work) address of the person."""
    email: Annotated[EmailStr | None, BeforeValidator(_unpack_variable)] = None
    """Email address of the person."""
    affiliation: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Affiliation of the person."""


class Software(BaseModel):
    """A piece of software.

    The piece of software should be specified as precisely as possible.
    For example, a release version of ScippNeutron could be specified as follows:

    >>> from scippneutron.metadata import Software
    >>> software = Software(
    ...     name='ScippNeutron',
    ...     version='24.11.0',
    ...     url='https://github.com/scipp/scippneutron/releases/tag/24.11.0',
    ...     doi='https://doi.org/10.5281/zenodo.14139599',
    ... )

    A development version might include a Git hash in the version.
    Alternative methods can be used for projects that do not use Git.
    But the software should be specified as precisely as possible.
    For example:

    >>> software = Software(
    ...     name='ScippNeutron',
    ...     version='24.11.1.dev8+g10d09ab0',
    ...     url='https://github.com/scipp/scippneutron',
    ... )

    The URL can either point to the source code, a release artifact, or a package
    index, such as ``pypi.org`` or ``anaconda.org``.
    """

    name: str
    """Name of the piece of software."""
    version: str
    """Complete version of the piece of software."""
    url: str | None = None
    """URL to the concrete version of the software.

    If no URL for a concrete version is available,
    a URL of the project or source code may be used.
    """
    doi: str | None = None
    """DOI of the concrete version of the software.

    If there is no DOI for the concrete version,
    a general DOI for the software may be used.
    """

    @classmethod
    def from_package_metadata(cls, package_name: str) -> Software:
        """Construct a Software instance from the metadata of an installed package.

        This function attempts to deduce all information it can from package metadata.
        But it only has access to the information that is encoded in the package.
        It therefore returns the base project URL instead of a concrete release URL,
        and it does not return a DOI.

        Parameters
        ----------
        package_name:
            The name of the Python package.

        Returns
        -------
        :
            A Software instance.
        """
        return cls(
            name=package_name,
            version=_deduce_package_version(package_name),
            url=_deduce_package_source_url(package_name),
            doi=None,
        )

    @property
    def name_version(self) -> str:
        """The name and version of the software, separated by a space."""
        return f'{self.name} {self.version}'

    @property
    def compact_repr(self) -> str:
        """A representation of this software as a single short string."""
        if self.url:
            return f'{self.name_version} ({self.url})'
        return self.name_version


def _deduce_package_version(package_name: str) -> str | None:
    from importlib.metadata import PackageNotFoundError, version

    try:
        return version(package_name)
    except PackageNotFoundError:
        # Either the package is not installed or has no metadata.
        from importlib import import_module

        try:
            package = import_module(package_name)
        except ModuleNotFoundError as e:
            raise e from None

        try:
            return package.__version__
        except AttributeError:
            raise RuntimeError(
                f"Package '{package_name}' has no metadata and no "
                f"__version__ attribute. Specify the version manually."
            ) from None


def _deduce_package_source_url(package_name: str) -> str | None:
    from importlib.metadata import PackageNotFoundError, metadata

    try:
        meta = metadata(package_name)
    except PackageNotFoundError:
        # Either the package is not installed or has no metadata.
        return None

    if not (urls := meta.get_all("project-url")):
        return None

    try:
        return next(
            url.split(',')[-1].strip() for url in urls if url.startswith("Source")
        )
    except StopIteration:
        return None


class SourceType(enum.Enum):
    """Type of source.

    Names are based on NeXus definitions.
    """

    SpallationNeutronSource = 'Spallation Neutron Source'
    """A spallation neutron source."""
    ReactorNeutronSource = 'Reactor Neutron Source'
    """A reactor neutron source."""
    SynchrotronXraySource = 'Synchrotron X-ray Source'
    """A synchrotron X-ray source."""


class RadiationProbe(enum.Enum):
    """Type of radiation probe.

    Names are based on NeXus definitions.
    """

    Neutron = 'neutron'
    """Neutron probe."""
    Xray = 'X-ray'
    """X-ray probe."""


class Source(BaseModel):
    """Information about a neutron source.

    The ESS source is provided as ``scippneutron.metadata.ESS_SOURCE``.
    """

    name: Annotated[str | None, BeforeValidator(_unpack_variable)] = None
    """Name of the source."""
    source_type: SourceType
    """Type of this source."""
    probe: RadiationProbe
    """Radiation probe of the source."""


ESS_SOURCE = Source(
    name="ESS Butterfly",
    source_type=SourceType.SpallationNeutronSource,
    probe=RadiationProbe.Neutron,
)
ESS_SOURCE.__doc__ = """Default parameters of the ESS source."""


def _read_optional_nexus_string(group: snx.Group | None, key: str) -> str | None:
    if group is None:
        return None
    if (ds := group.get(key)) is not None:
        return ds[()]
    return None


def _read_optional_nexus_datetime(group: snx.Group | None, key: str) -> datetime | None:
    if (s := _read_optional_nexus_string(group, key)) is not None:
        return parse_datetime(s)
    return None


def _get_unique_nexus_child(
    entry: snx.Group, nx_class: type, name: str | None
) -> snx.Group | None:
    if name is not None:
        return entry.get(name)
    children = entry[nx_class]
    if len(children) > 1:
        raise RuntimeError(
            f"Got multiple {nx_class.__name__} in NeXus entry '{entry.name}'"
        )
    if len(children) == 0:
        return None
    return next(iter(children.values()))


# More instruments may be added as needed.
# All instrument names are lowercase.
_FACILITY_PER_INSTRUMENT: dict[str, str | tuple[str, str]] = {
    # ESS
    'beer': 'ESS',
    'bifrost': 'ESS',
    'cspec': 'ESS',
    'dream': 'ESS',
    'estia': 'ESS',
    'freia': 'ESS',
    'heimdal': 'ESS',
    'loki': 'ESS',
    'magic': 'ESS',
    'miracles': 'ESS',
    'nmx': 'ESS',
    'odin': 'ESS',
    'skadi': 'ESS',
    'tbl': 'ESS',
    'trex': 'ESS',
    'vespa': 'ESS',
    # SINQ
    'amor': ('SINQ', 'PSI'),
}


# NeXus provides no way to specify the facility.
# But we can usually guess it based on the instrument name.
def _guess_facility_and_site(
    instrument_name: str | None,
) -> tuple[str | None, str | None]:
    if instrument_name is None:
        return None, None

    match _FACILITY_PER_INSTRUMENT.get(instrument_name.lower()):
        case None:
            return None, None
        case (facility, site):
            return facility, site
        case facility:
            return facility, facility



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from .conversions import (  # noqa: F401
    conversion_graph,
    convert,
    deduce_conversion_graph,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""Common file handling."""

from contextlib import AbstractContextManager, nullcontext
from io import BytesIO, StringIO
from os import PathLike
from typing import BinaryIO, Literal, TextIO, overload


@overload
def open_or_pass(
    path: str | PathLike[str] | TextIO,
    mode: Literal["r", "w", "r+"],
) -> AbstractContextManager[TextIO]: ...


@overload
def open_or_pass(
    path: StringIO,
    mode: Literal["r", "w", "r+"],
) -> AbstractContextManager[StringIO]: ...


@overload
def open_or_pass(
    path: str | PathLike[str] | BinaryIO | BytesIO,
    mode: Literal["rb", "wb", "r+b"],
) -> AbstractContextManager[BinaryIO]: ...


@overload
def open_or_pass(
    path: BytesIO,
    mode: Literal["rb", "wb", "r+b"],
) -> AbstractContextManager[BytesIO]: ...


def open_or_pass(
    path: str | PathLike[str] | TextIO | StringIO | BinaryIO | BytesIO,
    mode: Literal["r", "w", "r+", "rb", "wb", "r+b"],
) -> AbstractContextManager[TextIO | BinaryIO]:
    """Open a file at a path or return an already open file."""
    if isinstance(path, TextIO | StringIO | BytesIO | BinaryIO):
        return nullcontext(path)
    return open(path, mode)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import dataclasses
import os
from datetime import datetime, timezone
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Any, BinaryIO

import numpy as np
import scipp as sc

from .._files import open_or_pass
from . import _ir as ir
from ._bytes import Byteorder
from ._low_level_io import LowLevelSqw
from ._models import (
    DataBlockName,
    SqwDataBlockDescriptor,
    SqwDataBlockType,
    SqwDndMetadata,
    SqwFileHeader,
    SqwFileType,
    SqwIXExperiment,
    SqwIXNullInstrument,
    SqwIXSample,
    SqwMainHeader,
    SqwMultiIXExperiment,
    SqwPixelMetadata,
    UniqueObjContainer,
    UniqueRefContainer,
)
from ._read_write import write_object_array

# Based on
# https://github.com/pace-neutrons/Horace/blob/master/documentation/add/05_file_formats.md
_DEFAULT_PIX_ROWS = (
    "u1",  # Coordinate 1 (h)
    "u2",  # Coordinate 2 (k)
    "u3",  # Coordinate 3 (l)
    "u4",  # Coordinate 1 (E)
    "irun",  # Run index in header block
    "idet",  # Detector group number
    "ien",  # Energy bin number
    "signal",  # Signal
    "error",  # Variance
)
_DEFAULT_PIX_ROW_UNITS = (
    '1/angstrom',
    '1/angstrom',
    '1/angstrom',
    'meV',
    None,
    None,
    None,
    'count',
    'count**2',
)


class SqwBuilder:
    def __init__(
        self,
        path: str | PathLike[str] | BinaryIO | BytesIO,
        title: str,
        *,
        byteorder: Byteorder,
    ) -> None:
        self._path = path
        self._stored_path = (
            None if isinstance(self._path, BinaryIO | BytesIO) else Path(self._path)
        )
        self._byteorder = byteorder
        self._n_dims = 0

        main_header = SqwMainHeader(
            full_filename=self._full_filename,
            title=title,
            # To be replaced when registering pixel data.
            nfiles=0,
            # To be replaced when writing the file.
            creation_date=datetime(1, 1, 1, tzinfo=timezone.utc),
        )
        self._data_blocks: dict[DataBlockName, Any] = {  # TODO type
            ("", "main_header"): main_header,
        }

        self._dnd_placeholder: _DndPlaceholder | None = None
        self._pix_wrap: _PixWrap | None = None
        self._instrument: SqwIXNullInstrument | None = None
        self._sample: SqwIXSample | None = None

    def create(self, *, chunk_size: int = 8192) -> Path | None:
        with open_or_pass(self._path, "wb") as f:
            sqw_io = LowLevelSqw(
                f,
                path=self._stored_path,
                byteorder=self._byteorder,
            )

            _write_file_header(sqw_io, self._make_file_header())

            block_buffers, block_descriptors = self._serialize_data_blocks()
            bat_buffer, block_descriptors = self._serialize_block_allocation_table(
                block_descriptors=block_descriptors,
                bat_offset=sqw_io.position,
            )
            sqw_io.write_raw(bat_buffer)
            for name, buffer in block_buffers.items():
                descriptor = block_descriptors[name]
                match descriptor.block_type:
                    case SqwDataBlockType.regular:
                        # Type guaranteed by _serialize_data_blocks
                        sqw_io.write_raw(buffer)  # type: ignore[arg-type]
                    case SqwDataBlockType.pix:
                        # Type guaranteed by _serialize_data_blocks
                        self._pix_wrap.write(sqw_io, chunk_size=chunk_size)  # type: ignore[union-attr]
                    case SqwDataBlockType.dnd:
                        # Type guaranteed by _serialize_data_blocks
                        self._dnd_placeholder.write(sqw_io)  # type: ignore[union-attr]
                    case _:
                        raise NotImplementedError(
                            f"Unsupported data block type: {descriptor.block_type}"
                        )

        return self._stored_path

    def add_pixel_data(
        self,
        data: sc.DataArray,
        *,
        experiments: list[SqwIXExperiment],
        n_dims: int = 4,
        rows: tuple[str, ...] = _DEFAULT_PIX_ROWS,
        row_units: tuple[str | None, ...] = _DEFAULT_PIX_ROW_UNITS,
    ) -> SqwBuilder:
        self._n_dims = n_dims
        self._data_blocks[("experiment_info", "expdata")] = SqwMultiIXExperiment(
            experiments
        )

        self._pix_wrap = _split_pix_rows(data, rows, row_units)
        metadata = self._make_pix_metadata(self._pix_wrap)
        self._data_blocks[("pix", "metadata")] = metadata
        self._data_blocks[("", "main_header")].nfiles = len(experiments)
        return self

    def add_empty_detector_params(self) -> SqwBuilder:
        self._data_blocks[("", "detpar")] = UniqueRefContainer(
            global_name="GLOBAL_NAME_DETECTORS_CONTAINER",
            objects=UniqueObjContainer(
                baseclass="IX_detector_array",
                objects=[],
                indices=[],
            ),
        )
        return self

    def _add_dnd_metadata(self, block: SqwDndMetadata) -> SqwBuilder:
        self._data_blocks[("data", "metadata")] = block
        return self

    def add_empty_dnd_data(self, block: SqwDndMetadata) -> SqwBuilder:
        # The file must always contain a DND block
        builder = self._add_dnd_metadata(block)
        builder._dnd_placeholder = _DndPlaceholder(
            shape=tuple(map(int, block.axes.n_bins_all_dims))  # type: ignore[call-overload]
        )
        return builder

    def add_default_instrument(self, instrument: SqwIXNullInstrument) -> SqwBuilder:
        self._instrument = instrument
        return self

    def add_default_sample(self, sample: SqwIXSample) -> SqwBuilder:
        self._sample = sample
        return self

    def _make_file_header(self) -> SqwFileHeader:
        return SqwFileHeader(
            prog_name="horace",
            prog_version=4.0,
            sqw_type=SqwFileType.SQW,
            n_dims=self._n_dims,
        )

    def _make_pix_metadata(self, pix_wrap: _PixWrap) -> SqwPixelMetadata:
        return SqwPixelMetadata(
            full_filename=self._full_filename,
            npix=pix_wrap.n_pixels(),
            data_range=np.vstack(
                [
                    (
                        sc.to_unit(row.min(), unit).value,
                        sc.to_unit(row.max(), unit).value,
                    )
                    for row, unit in zip(
                        pix_wrap.row_data, pix_wrap.row_units, strict=True
                    )
                ]
            ),
        )

    def _serialize_data_blocks(
        self,
    ) -> tuple[
        dict[DataBlockName, memoryview | None],
        dict[DataBlockName, SqwDataBlockDescriptor],
    ]:
        data_blocks = self._prepare_data_blocks()
        buffers: dict[DataBlockName, memoryview | None] = {}
        descriptors = {}
        for name, data_block in data_blocks.items():
            buffer = BytesIO()
            sqw_io = LowLevelSqw(
                buffer, path=self._stored_path, byteorder=self._byteorder
            )
            write_object_array(sqw_io, data_block.serialize_to_ir().to_object_array())

            buffer.seek(0)
            buf = buffer.getbuffer()
            buffers[name] = buf
            descriptors[name] = SqwDataBlockDescriptor(
                block_type=SqwDataBlockType.regular,
                name=name,
                position=0,
                size=len(buf),
                locked=False,
            )

        if self._dnd_placeholder is not None:
            buffers[("data", "nd_data")] = None
            descriptors[("data", "nd_data")] = SqwDataBlockDescriptor(
                block_type=SqwDataBlockType.dnd,
                name=("data", "nd_data"),
                position=0,
                size=self._dnd_placeholder.size(),
                locked=False,
            )

        if self._pix_wrap is not None:
            buffers[("pix", "data_wrap")] = None
            descriptors[("pix", "data_wrap")] = SqwDataBlockDescriptor(
                block_type=SqwDataBlockType.pix,
                name=("pix", "data_wrap"),
                position=0,
                size=self._pix_wrap.size(),
                locked=False,
            )

        return buffers, descriptors

    def _prepare_data_blocks(self) -> dict[DataBlockName, Any]:
        filepath, filename = self._filepath_and_name
        blocks = {
            key: block.prepare_for_serialization(filepath=filepath, filename=filename)
            for key, block in self._data_blocks.items()
        }

        nfiles = blocks[("", "main_header")].nfiles
        if self._instrument is not None:
            blocks[("experiment_info", "instruments")] = _broadcast_unique_ref(
                self._instrument,
                n=nfiles,
                baseclass="IX_inst",
                global_name="GLOBAL_NAME_INSTRUMENTS_CONTAINER",
            )
        if self._sample is not None:
            blocks[("experiment_info", "samples")] = _broadcast_unique_ref(
                self._sample,
                n=nfiles,
                baseclass="IX_samp",
                global_name="GLOBAL_NAME_SAMPLES_CONTAINER",
            )

        return _to_canonical_block_order(blocks)

    def _serialize_block_allocation_table(
        self,
        block_descriptors: dict[DataBlockName, SqwDataBlockDescriptor],
        bat_offset: int,
    ) -> tuple[memoryview, dict[DataBlockName, SqwDataBlockDescriptor]]:
        # This function first writes the block allocation table (BAT) with placeholder
        # values in order to determine the size of the BAT.
        # Then, it computes the actual positions that data blocks will have in the file
        # and inserts those positions into the serialized BAT.
        # It returns a buffer of the BAT that can be inserted right after the file
        # header and an updated in-memory representation of the BAT.

        buffer = BytesIO()
        sqw_io = LowLevelSqw(buffer, path=self._stored_path, byteorder=self._byteorder)
        sqw_io.write_u32(0)  # Size of BAT in bytes, filled in below.
        bat_begin = sqw_io.position
        sqw_io.write_u32(len(block_descriptors))
        # Offsets are relative to the local sqw_io.
        position_offsets = {
            name: _write_data_block_descriptor(sqw_io, descriptor)
            for name, descriptor in block_descriptors.items()
        }
        bat_size = sqw_io.position - bat_begin

        block_position = bat_offset + sqw_io.position
        amended_descriptors = {}
        for name, descriptor in block_descriptors.items():
            amended_descriptors[name] = dataclasses.replace(
                descriptor, position=block_position
            )
            offset = position_offsets[name]
            sqw_io.seek(offset)
            sqw_io.write_u64(block_position)
            block_position += descriptor.size

        sqw_io.seek(0)
        sqw_io.write_u32(bat_size)
        return buffer.getbuffer(), amended_descriptors

    @property
    def _full_filename(self) -> str:
        return os.fspath(self._stored_path or "in_memory")

    @property
    def _filepath_and_name(self) -> tuple[str, str]:
        if self._stored_path is None:
            return "", ""
        return os.fspath(self._stored_path.parent), self._stored_path.name


def _write_file_header(sqw_io: LowLevelSqw, file_header: SqwFileHeader) -> None:
    sqw_io.write_char_array(file_header.prog_name)
    sqw_io.write_f64(file_header.prog_version)
    sqw_io.write_u32(file_header.sqw_type.value)
    sqw_io.write_u32(file_header.n_dims)


def _write_data_block_descriptor(
    sqw_io: LowLevelSqw, descriptor: SqwDataBlockDescriptor
) -> int:
    sqw_io.write_char_array(descriptor.block_type.value)
    sqw_io.write_char_array(descriptor.name[0])
    sqw_io.write_char_array(descriptor.name[1])
    pos = sqw_io.position
    sqw_io.write_u64(descriptor.position)
    sqw_io.write_u32(descriptor.size)
    sqw_io.write_u32(int(descriptor.locked))
    return pos


def _broadcast_unique_ref(
    obj: ir.Serializable,
    n: int,
    baseclass: str,
    global_name: str,
) -> UniqueRefContainer:
    return UniqueRefContainer(
        global_name=global_name,
        objects=UniqueObjContainer(
            baseclass=baseclass,
            objects=[obj],
            indices=[0] * n,
        ),
    )


def _to_canonical_block_order(
    blocks: dict[DataBlockName, Any],
) -> dict[DataBlockName, Any]:
    order = (
        ("", "main_header"),
        ("", "detpar"),
        ("data", "metadata"),
        ("data", "nd_data"),
        ("experiment_info", "instruments"),
        ("experiment_info", "samples"),
        ("experiment_info", "expdata"),
        ("pix", "metadata"),
        ("pix", "data_wrap"),
    )
    blocks = dict(blocks)
    out = {name: block for name in order if (block := blocks.get(name)) is not None}
    out.update(blocks)  # append remaining blocks if any
    return out


def _split_pix_rows(
    data: sc.DataArray, rows: tuple[str, ...], row_units: tuple[str | None, ...]
) -> _PixWrap:
    """Prepare the selected pixel rows for writing."""
    selected = []
    for name in rows:
        if name == 'signal':
            selected.append(sc.values(data.data))
        elif name == 'error':
            selected.append(sc.variances(data.data))
        else:
            if data.coords.is_edges(name, data.dim):
                raise sc.BinEdgeError(
                    f"Pixel data must not contain bin-edges, got edges for '{name}'."
                )
            selected.append(data.coords[name])
    return _PixWrap(
        row_data=selected,
        row_units=row_units,
    )


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class _DndPlaceholder:
    shape: tuple[int, ...]

    def size(self) -> int:
        n_elem = int(np.prod(self.shape))
        return 4 + 4 * len(self.shape) + 3 * 8 * n_elem

    def write(self, sqw_io: LowLevelSqw) -> None:
        sqw_io.write_u32(len(self.shape))
        for s in self.shape:
            sqw_io.write_u32(s)
        zero = np.zeros(self.shape, dtype="float64")
        sqw_io.write_array(zero)
        sqw_io.write_array(zero)
        sqw_io.write_array(np.zeros(self.shape, dtype="uint64"))


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class _PixWrap:
    row_data: list[sc.Variable]
    row_units: tuple[str | None, ...]

    def size(self) -> int:
        # *4 for f32
        return 4 + 8 + self.n_rows() * 4 * self.n_pixels()

    def n_rows(self) -> int:
        return len(self.row_data)

    def n_pixels(self) -> int:
        return len(self.row_data[0])

    def write(self, sqw_io: LowLevelSqw, chunk_size: int) -> None:
        sqw_io.write_u32(self.n_rows())
        sqw_io.write_u64(self.n_pixels())

        buffer = np.empty((self.n_pixels(), self.n_rows()), dtype=np.float32)
        remaining = self.n_pixels()
        for offset in range(0, self.n_rows(), chunk_size):
            n = min(chunk_size, remaining)
            remaining -= n
            for i_row, (row, unit) in enumerate(
                zip(self.row_data, self.row_units, strict=True)
            ):
                buffer[:n, i_row] = sc.to_unit(
                    row[offset : offset + chunk_size], unit, copy=False
                ).values
            sqw_io.write_array(buffer[:n])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401, I

"""Neutron scattering toolkit built using scipp for Data Reduction.

ScippNeutron is a generic (as in 'usable by different facilities')
package for data processing in neutron scattering.
It provides coordinate transformations, file I/O, and technique-specific tools.

See the online documentation for user guides and the API reference:
https://scipp.github.io/scippneutron/
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

import lazy_loader as lazy

submodules = ['atoms' 'chopper', 'io', 'conversion']

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=submodules,
    submod_attrs={
        'beamline_components': [
            'position',
            'source_position',
            'sample_position',
            'incident_beam',
            'scattered_beam',
            'Ltotal',
            'L1',
            'L2',
            'two_theta',
        ],
        'core': ['convert', 'conversion_graph', 'deduce_conversion_graph'],
        'mantid': ['from_mantid', 'to_mantid', 'load_with_mantid', 'fit'],
        'instrument_view': ['instrument_view'],
        'masking': ['MaskingTool'],
    },
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from .disk_chopper import DiskChopper, DiskChopperType  # noqa: F401
from .filtering import collapse_plateaus, filter_in_phase, find_plateaus  # noqa: F401
from .nexus_chopper import extract_chopper_from_nexus  # noqa: F401



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import dataclasses
import sys
from datetime import datetime, timedelta, timezone
from io import BytesIO
from pathlib import Path
from typing import Literal

import numpy as np
import pytest
import scipp as sc
import scipp.testing

from scippneutron.io.sqw import (
    Byteorder,
    EnergyMode,
    Sqw,
    SqwDndMetadata,
    SqwIXExperiment,
    SqwLineAxes,
    SqwLineProj,
)


class _PathBuffer:
    def __init__(self, path: Path) -> None:
        self.path = path

    def get(self) -> Path:
        return self.path

    def rewind(self) -> None:
        pass

    def read(self, n: int) -> bytes:
        with self.path.open("rb") as file:
            return file.read(n)


class _BytesBuffer:
    def __init__(self) -> None:
        self.buffer = BytesIO()

    def get(self) -> BytesIO:
        return self.buffer

    def rewind(self) -> None:
        self.buffer.seek(0)

    def read(self, n: int) -> bytes:
        return self.buffer.read(n)


@pytest.fixture(params=["buffer", "file"])
def buffer(
    request: pytest.FixtureRequest, tmp_path: Path
) -> _BytesBuffer | _PathBuffer:
    match request.param:
        case "buffer":
            return _BytesBuffer()
        case "file":
            return _PathBuffer(tmp_path / "sqw_file.sqw")


def test_create_sets_byteorder_native(buffer: _BytesBuffer | _PathBuffer) -> None:
    builder = Sqw.build(buffer.get())
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        assert sqw.byteorder.value == sys.byteorder


def test_create_sets_byteorder_little(buffer: _BytesBuffer | _PathBuffer) -> None:
    builder = Sqw.build(buffer.get(), byteorder="little")
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        assert sqw.byteorder == Byteorder.little


def test_create_sets_byteorder_big(buffer: _BytesBuffer | _PathBuffer) -> None:
    builder = Sqw.build(buffer.get(), byteorder="big")
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        assert sqw.byteorder == Byteorder.big


def test_create_writes_file_header_little_endian(
    buffer: _BytesBuffer | _PathBuffer,
) -> None:
    builder = Sqw.build(buffer.get(), byteorder="little")
    builder.create()

    buffer.rewind()
    expected = (
        b"\x06\x00\x00\x00"
        b"horace"
        b"\x00\x00\x00\x00\x00\x00\x10\x40"
        b"\x01\x00\x00\x00"
        b"\x00\x00\x00\x00"
    )
    assert buffer.read(len(expected)) == expected


def test_create_writes_file_header_big_endian(
    buffer: _BytesBuffer | _PathBuffer,
) -> None:
    builder = Sqw.build(buffer.get(), byteorder="big")
    builder.create()

    buffer.rewind()
    expected = (
        b"\x00\x00\x00\x06"
        b"horace"
        b"\x40\x10\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x01"
        b"\x00\x00\x00\x00"
    )
    assert buffer.read(len(expected)) == expected


@pytest.mark.parametrize("byteorder", ["native", "little", "big"])
def test_create_writes_main_header(
    byteorder: Literal["native", "little", "big"], buffer: _BytesBuffer | _PathBuffer
) -> None:
    builder = Sqw.build(buffer.get(), title="my title", byteorder=byteorder)
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        main_header = sqw.read_data_block(("", "main_header"))
    filename = "in_memory" if isinstance(buffer, _BytesBuffer) else str(buffer.get())
    assert main_header.full_filename == filename
    assert main_header.title == "my title"
    assert main_header.nfiles == 0
    assert (main_header.creation_date - datetime.now(tz=timezone.utc)) < timedelta(
        seconds=1
    )


@pytest.mark.parametrize("byteorder", ["native", "little", "big"])
def test_writes_expdata(
    byteorder: Literal["native", "little", "big"], buffer: _BytesBuffer | _PathBuffer
) -> None:
    experiments = [
        SqwIXExperiment(
            run_id=0,
            efix=sc.scalar(1.2, unit="meV"),
            emode=EnergyMode.direct,
            en=sc.array(dims=["energy_transfer"], values=[3.0], unit="ueV"),
            psi=sc.scalar(1.2, unit="rad"),
            u=sc.vector([0.0, 1.0, 0.0]),
            v=sc.vector([1.0, 1.0, 0.0]),
            omega=sc.scalar(1.4, unit="rad"),
            dpsi=sc.scalar(46, unit="deg"),
            gl=sc.scalar(3, unit="rad"),
            gs=sc.scalar(-0.5, unit="rad"),
            filename="run1.nxspe",
            filepath="/data",
        ),
        SqwIXExperiment(
            run_id=2,
            efix=sc.scalar(0.16, unit="eV"),
            emode=EnergyMode.direct,
            en=sc.array(dims=["energy_transfer"], values=[2.0, 4.5], unit="meV"),
            psi=sc.scalar(-10.0, unit="deg"),
            u=sc.vector([1.0, 0.0, 0.0]),
            v=sc.vector([0.0, 1.0, 0.0]),
            omega=sc.scalar(-91, unit="deg"),
            dpsi=sc.scalar(-0.5, unit="rad"),
            gl=sc.scalar(0.0, unit="deg"),
            gs=sc.scalar(-5, unit="deg"),
            filename="run2.nxspe",
            filepath="/data",
        ),
    ]
    # The same as above but with canonical units.
    expected_experiments = [
        SqwIXExperiment(
            run_id=0,
            efix=sc.scalar(1.2, unit="meV"),
            emode=EnergyMode.direct,
            en=sc.array(dims=["energy_transfer"], values=[0.003], unit="meV"),
            psi=sc.scalar(1.2, unit="rad"),
            u=sc.vector([0.0, 1.0, 0.0]),
            v=sc.vector([1.0, 1.0, 0.0]),
            omega=sc.scalar(1.4, unit="rad"),
            dpsi=sc.scalar(46.0, unit="deg").to(unit="rad"),
            gl=sc.scalar(3.0, unit="rad"),
            gs=sc.scalar(-0.5, unit="rad"),
            filename="run1.nxspe",
            filepath="/data",
        ),
        SqwIXExperiment(
            run_id=2,
            efix=sc.scalar(160.0, unit="meV"),
            emode=EnergyMode.direct,
            en=sc.array(dims=["energy_transfer"], values=[2.0, 4.5], unit="meV"),
            psi=sc.scalar(-10.0, unit="deg").to(unit="rad"),
            u=sc.vector([1.0, 0.0, 0.0]),
            v=sc.vector([0.0, 1.0, 0.0]),
            omega=sc.scalar(-91.0, unit="deg").to(unit="rad"),
            dpsi=sc.scalar(-0.5, unit="rad"),
            gl=sc.scalar(0.0, unit="deg").to(unit="rad"),
            gs=sc.scalar(-5.0, unit="deg").to(unit="rad"),
            filename="run2.nxspe",
            filepath="/data",
        ),
    ]

    n_pixels = 7
    pixels = sc.DataArray(
        sc.zeros(sizes={'obs': n_pixels}, with_variances=True, unit='count'),
        coords={
            'u1': sc.zeros(sizes={'obs': n_pixels}, unit='1/Å'),
            'u2': sc.zeros(sizes={'obs': n_pixels}, unit='1/Å'),
            'u3': sc.zeros(sizes={'obs': n_pixels}, unit='1/Å'),
            'u4': sc.zeros(sizes={'obs': n_pixels}, unit='meV'),
            'idet': sc.zeros(sizes={'obs': n_pixels}, dtype=int, unit=None),
            'irun': sc.zeros(sizes={'obs': n_pixels}, dtype=int, unit=None),
            'ien': sc.zeros(sizes={'obs': n_pixels}, dtype=int, unit=None),
        },
    )

    builder = Sqw.build(buffer.get(), byteorder=byteorder)
    builder = builder.add_pixel_data(pixels, experiments=experiments)
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        loaded_experiments = sqw.read_data_block(("experiment_info", "expdata"))

    for loaded, expected in zip(loaded_experiments, expected_experiments, strict=True):
        for field in dataclasses.fields(expected):
            sc.testing.assert_identical(
                getattr(loaded, field.name), getattr(expected, field.name)
            )


@pytest.mark.parametrize("byteorder", ["native", "little", "big"])
def test_writes_pixel_data(
    byteorder: Literal["native", "little", "big"], buffer: _BytesBuffer | _PathBuffer
) -> None:
    rng = np.random.default_rng(1732)
    experiment_template = SqwIXExperiment(
        run_id=-1,
        efix=sc.scalar(1.2, unit="meV"),
        emode=EnergyMode.direct,
        en=sc.array(dims=["energy_transfer"], values=[3.0], unit="meV"),
        psi=sc.scalar(1.2, unit="rad"),
        u=sc.vector([0.0, 1.0, 0.0]),
        v=sc.vector([1.0, 1.0, 0.0]),
        omega=sc.scalar(1.4, unit="rad"),
        dpsi=sc.scalar(0.0, unit="rad"),
        gl=sc.scalar(3, unit="rad"),
        gs=sc.scalar(-0.5, unit="rad"),
        filename="",
        filepath="/data",
    )
    experiments = [
        dataclasses.replace(experiment_template, run_id=0, filename="f1"),
        dataclasses.replace(experiment_template, run_id=1, filename="f2"),
    ]

    # Chosen numbers can all be represented in float32 to allow exact comparisons.
    n_pixels = 7
    original = sc.DataArray(
        sc.array(
            dims=['obs'],
            values=rng.uniform(0, 100, n_pixels).astype('float32'),
            variances=rng.uniform(0.1, 1, n_pixels).astype('float32'),
            unit='count',
        ),
        coords={
            'idet': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(3),
            'irun': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(2),
            'ien': sc.arange('obs', 0, 2 * n_pixels, 2, unit=None).astype(int)
            // sc.index(10),
            'u1': sc.arange('obs', 0.0, n_pixels + 0.0, unit='1/Å'),
            'u2': sc.arange('obs', 1.0, n_pixels + 1.0, unit='1/Å'),
            'u3': sc.arange('obs', 2.0, n_pixels + 2.0, unit='1/Å'),
            'u4': sc.arange('obs', n_pixels, unit='meV') * 2,
        },
    )
    builder = Sqw.build(buffer.get(), byteorder=byteorder)
    builder = builder.add_pixel_data(original, experiments=experiments)
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        loaded = sqw.read_data_block(("pix", "data_wrap"))

    np.testing.assert_equal(loaded[:, 0], original.coords['u1'].values)
    np.testing.assert_equal(loaded[:, 1], original.coords['u2'].values)
    np.testing.assert_equal(loaded[:, 2], original.coords['u3'].values)
    np.testing.assert_equal(loaded[:, 3], original.coords['u4'].values)
    np.testing.assert_equal(loaded[:, 4], original.coords['irun'].values)
    np.testing.assert_equal(loaded[:, 5], original.coords['idet'].values)
    np.testing.assert_equal(loaded[:, 6], original.coords['ien'].values)
    np.testing.assert_equal(loaded[:, 7], original.values)
    np.testing.assert_equal(loaded[:, 8], original.variances)


@pytest.mark.parametrize("byteorder", ["native", "little", "big"])
def test_writes_pixel_data_chunked(
    byteorder: Literal["native", "little", "big"], buffer: _BytesBuffer | _PathBuffer
) -> None:
    rng = np.random.default_rng(1732)
    experiment_template = SqwIXExperiment(
        run_id=-1,
        efix=sc.scalar(1.2, unit="meV"),
        emode=EnergyMode.direct,
        en=sc.array(dims=["energy_transfer"], values=[3.0], unit="meV"),
        psi=sc.scalar(1.2, unit="rad"),
        u=sc.vector([0.0, 1.0, 0.0]),
        v=sc.vector([1.0, 1.0, 0.0]),
        omega=sc.scalar(1.4, unit="rad"),
        dpsi=sc.scalar(0.0, unit="rad"),
        gl=sc.scalar(3, unit="rad"),
        gs=sc.scalar(-0.5, unit="rad"),
        filename="",
        filepath="/data",
    )
    experiments = [
        dataclasses.replace(experiment_template, run_id=0, filename="f1"),
        dataclasses.replace(experiment_template, run_id=1, filename="f2"),
    ]

    # Chosen numbers can all be represented in float32 to allow exact comparisons.
    n_pixels = 7
    original = sc.DataArray(
        sc.array(
            dims=['obs'],
            values=rng.uniform(0, 100, n_pixels).astype('float32'),
            variances=rng.uniform(0.1, 1, n_pixels).astype('float32'),
            unit='count',
        ),
        coords={
            'idet': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(3),
            'irun': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(2),
            'ien': sc.arange('obs', 0, 2 * n_pixels, 2, unit=None).astype(int)
            // sc.index(10),
            'u1': sc.arange('obs', 0.0, n_pixels + 0.0, unit='1/Å'),
            'u2': sc.arange('obs', 1.0, n_pixels + 1.0, unit='1/Å'),
            'u3': sc.arange('obs', 2.0, n_pixels + 2.0, unit='1/Å'),
            'u4': sc.arange('obs', n_pixels, unit='meV') * 2,
        },
    )
    builder = Sqw.build(buffer.get(), byteorder=byteorder)
    builder = builder.add_pixel_data(original, experiments=experiments)
    builder.create(chunk_size=n_pixels // 3)
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        loaded = sqw.read_data_block(("pix", "data_wrap"))

    np.testing.assert_equal(loaded[:, 0], original.coords['u1'].values)
    np.testing.assert_equal(loaded[:, 1], original.coords['u2'].values)
    np.testing.assert_equal(loaded[:, 2], original.coords['u3'].values)
    np.testing.assert_equal(loaded[:, 3], original.coords['u4'].values)
    np.testing.assert_equal(loaded[:, 4], original.coords['irun'].values)
    np.testing.assert_equal(loaded[:, 5], original.coords['idet'].values)
    np.testing.assert_equal(loaded[:, 6], original.coords['ien'].values)
    np.testing.assert_equal(loaded[:, 7], original.values)
    np.testing.assert_equal(loaded[:, 8], original.variances)


@pytest.mark.parametrize("byteorder", ["native", "little", "big"])
def test_writes_pixel_data_convert_units(
    byteorder: Literal["native", "little", "big"], buffer: _BytesBuffer | _PathBuffer
) -> None:
    rng = np.random.default_rng(1732)
    experiment_template = SqwIXExperiment(
        run_id=-1,
        efix=sc.scalar(1.2, unit="meV"),
        emode=EnergyMode.direct,
        en=sc.array(dims=["energy_transfer"], values=[3.0], unit="meV"),
        psi=sc.scalar(1.2, unit="rad"),
        u=sc.vector([0.0, 1.0, 0.0]),
        v=sc.vector([1.0, 1.0, 0.0]),
        omega=sc.scalar(1.4, unit="rad"),
        dpsi=sc.scalar(0.0, unit="rad"),
        gl=sc.scalar(3, unit="rad"),
        gs=sc.scalar(-0.5, unit="rad"),
        filename="",
        filepath="/data",
    )
    experiments = [
        dataclasses.replace(experiment_template, run_id=0, filename="f1"),
        dataclasses.replace(experiment_template, run_id=1, filename="f2"),
    ]

    n_pixels = 7
    original = sc.DataArray(
        sc.array(
            dims=['obs'],
            values=rng.uniform(0, 100, n_pixels),
            variances=rng.uniform(0.1, 1, n_pixels),
            unit='mega count',
        ),
        coords={
            'idet': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(3),
            'irun': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(2),
            'ien': sc.arange('obs', 0, 2 * n_pixels, 2, unit=None).astype(int)
            // sc.index(10),
            'u1': sc.arange('obs', 0.0, n_pixels + 0.0, unit='1/fm'),
            'u2': sc.arange('obs', 1.0, n_pixels + 1.0, unit='10/Å'),
            'u3': sc.arange('obs', 2.0, n_pixels + 2.0, unit='1/um'),
            'u4': sc.arange('obs', n_pixels, unit='eV') * 2,
        },
    )
    builder = Sqw.build(buffer.get(), byteorder=byteorder)
    builder = builder.add_pixel_data(original, experiments=experiments)
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        loaded = sqw.read_data_block(("pix", "data_wrap"))

    np.testing.assert_allclose(
        loaded[:, 0], original.coords['u1'].to(unit='1/Å').values
    )
    np.testing.assert_allclose(
        loaded[:, 1], original.coords['u2'].to(unit='1/Å').values
    )
    np.testing.assert_allclose(
        loaded[:, 2], original.coords['u3'].to(unit='1/Å').values
    )
    np.testing.assert_allclose(
        loaded[:, 3], original.coords['u4'].to(unit='meV').values
    )
    np.testing.assert_equal(loaded[:, 4], original.coords['irun'].values)
    np.testing.assert_equal(loaded[:, 5], original.coords['idet'].values)
    np.testing.assert_equal(loaded[:, 6], original.coords['ien'].values)
    np.testing.assert_allclose(loaded[:, 7], original.to(unit='count').values)
    np.testing.assert_allclose(loaded[:, 8], original.to(unit='count').variances)


@pytest.mark.parametrize("byteorder", ["native", "little", "big"])
def test_writes_data_metadata(
    byteorder: Literal["native", "little", "big"], buffer: _BytesBuffer | _PathBuffer
) -> None:
    metadata = SqwDndMetadata(
        axes=SqwLineAxes(
            title="test axes",
            label=["x", "y", "z", "dE"],
            img_scales=[
                1.0 / sc.Unit("angstrom"),
                2.0 / sc.Unit("2*angstrom"),
                0.5 / sc.Unit("kilo angstrom"),
                0.2 * sc.Unit("eV"),
            ],
            img_range=[
                sc.array(dims=["range"], values=[-30.0, 540.0], unit="1/(k angstrom)"),
                sc.array(dims=["range"], values=[-0.5, 6.7], unit="1/angstrom"),
                sc.array(dims=["range"], values=[-5.6, -2.4], unit="10/angstrom"),
                sc.array(dims=["range"], values=[6.0, 9.1], unit="meV"),
            ],
            n_bins_all_dims=sc.array(dims=["axis"], values=[40, 50, 40, 40], unit=None),
            single_bin_defines_iax=sc.array(
                dims=["axis"], values=[False, True, True, True]
            ),
            dax=sc.array(dims=["axis"], values=[2, 1, 0, 3], unit=None),
            offset=[
                1.0 / sc.Unit("2*angstrom"),
                50.0 / sc.Unit("milli angstrom"),
                0.0 / sc.Unit("angstrom"),
                0.0 * sc.Unit("meV"),
            ],
            changes_aspect_ratio=True,
        ),
        proj=SqwLineProj(
            lattice_spacing=sc.vector([2.1, 2.1, 2.5], unit="angstrom"),
            lattice_angle=sc.vector([np.pi / 2, np.pi / 4, np.pi / 2], unit="rad"),
            offset=[
                1.0 / sc.Unit("2*angstrom"),
                50.0 / sc.Unit("milli angstrom"),
                0.0 / sc.Unit("angstrom"),
                0.0 * sc.Unit("meV"),
            ],
            title="my projection",
            label=["x", "y", "z", "dE"],
            u=sc.vector([0.0, 1.0, 0.0], unit="1/angstrom"),
            v=sc.vector([1.0, 0.0, 0.0], unit="1/(milli angstrom)"),
            w=None,
            non_orthogonal=False,
            type="aaa",
        ),
    )
    # The same as above but with canonical units.
    expected_metadata = SqwDndMetadata(
        axes=SqwLineAxes(
            title="test axes",
            label=["x", "y", "z", "dE"],
            img_scales=[
                1.0 / sc.Unit("angstrom"),
                1.0 / sc.Unit("angstrom"),
                0.0005 / sc.Unit("angstrom"),
                200.0 * sc.Unit("meV"),
            ],
            img_range=[
                sc.array(dims=["range"], values=[-0.03, 0.540], unit="1/angstrom"),
                sc.array(dims=["range"], values=[-0.5, 6.7], unit="1/angstrom"),
                sc.array(dims=["range"], values=[-56.0, -24.0], unit="1/angstrom"),
                sc.array(dims=["range"], values=[6.0, 9.1], unit="meV"),
            ],
            n_bins_all_dims=sc.array(dims=["axis"], values=[40, 50, 40, 40], unit=None),
            single_bin_defines_iax=sc.array(
                dims=["axis"], values=[False, True, True, True]
            ),
            dax=sc.array(dims=["axis"], values=[2, 1, 0, 3], unit=None),
            offset=[
                0.5 / sc.Unit("angstrom"),
                50000.0 / sc.Unit("angstrom"),
                0.0 / sc.Unit("angstrom"),
                0.0 * sc.Unit("meV"),
            ],
            changes_aspect_ratio=True,
            filename="" if isinstance(buffer, _BytesBuffer) else buffer.get().name,
            filepath=""
            if isinstance(buffer, _BytesBuffer)
            else str(buffer.get().parent),
        ),
        proj=SqwLineProj(
            lattice_spacing=sc.vector([2.1, 2.1, 2.5], unit="angstrom"),
            lattice_angle=sc.vector([90.0, 45.0, 90.0], unit="deg"),
            offset=[
                0.5 / sc.Unit("angstrom"),
                50000.0 / sc.Unit("angstrom"),
                0.0 / sc.Unit("angstrom"),
                0.0 * sc.Unit("meV"),
            ],
            title="my projection",
            label=["x", "y", "z", "dE"],
            u=sc.vector([0.0, 1.0, 0.0], unit="1/angstrom"),
            v=sc.vector([1000.0, 0.0, 0.0], unit="1/angstrom"),
            w=None,
            non_orthogonal=False,
            type="aaa",
        ),
    )

    builder = Sqw.build(buffer.get(), byteorder=byteorder)
    builder = builder.add_empty_dnd_data(metadata)
    builder.create()
    buffer.rewind()

    with Sqw.open(buffer.get()) as sqw:
        loaded_metadata = sqw.read_data_block(("data", "metadata"))

    loaded_axes = loaded_metadata.axes
    expected_axes = expected_metadata.axes
    for field in dataclasses.fields(loaded_axes):
        loaded = getattr(loaded_axes, field.name)
        expected = getattr(expected_axes, field.name)
        if isinstance(loaded, list):
            for a, b in zip(loaded, expected, strict=True):
                sc.testing.assert_identical(a, b)
        else:
            sc.testing.assert_identical(loaded, expected)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import functools
import struct
from collections.abc import Callable
from io import BytesIO
from pathlib import Path
from typing import BinaryIO, ParamSpec, TypeVar

import numpy as np
import numpy.typing as npt

from ._bytes import Byteorder

_P = ParamSpec("_P")
_R = TypeVar("_R")
_E = TypeVar("_E", bound=np.generic, covariant=True)


def _annotate_read_exception(
    ty: str,
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Add a note with file-information to exceptions from read_* functions."""

    def decorator(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @functools.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            try:
                return func(*args, **kwargs)
            except (ValueError, UnicodeDecodeError) as exc:
                sqw_io: LowLevelSqw = args[0]  # type: ignore[assignment]
                _add_note_to_read_exception(exc, sqw_io, ty)
                raise

        return wrapper

    return decorator


def _add_note_to_read_exception(exc: Exception, sqw_io: LowLevelSqw, ty: str) -> None:
    path_piece = (
        "in-memory SQW file" if sqw_io.path is None else f"SQW file '{sqw_io.path}'"
    )
    _add_note(
        exc,
        f"When reading a {ty} from {path_piece} at position {sqw_io.position}",
    )


def _annotate_write_exception(
    ty: str,
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Add a note with file-information to exceptions from write_* functions."""

    def decorator(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @functools.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            try:
                return func(*args, **kwargs)
            except (ValueError, UnicodeEncodeError, OverflowError) as exc:
                sqw_io: LowLevelSqw = args[0]  # type: ignore[assignment]
                _add_note_to_write_exception(exc, sqw_io, ty)
                raise

        return wrapper

    return decorator


def _add_note_to_write_exception(exc: Exception, sqw_io: LowLevelSqw, ty: str) -> None:
    path_piece = (
        "in-memory SQW file" if sqw_io.path is None else f"SQW file '{sqw_io.path}'"
    )
    _add_note(
        exc,
        f"When writing a {ty} to {path_piece} at position {sqw_io.position}",
    )


def _add_note(exc: Exception, note: str) -> None:
    try:
        exc.add_note(note)  # type: ignore[attr-defined]
    except AttributeError:
        # Python < 3.11 -> do nothing and accept throwing a worse error.
        pass


class LowLevelSqw:
    def __init__(
        self, file: BinaryIO, *, path: Path | None, byteorder: Byteorder | None = None
    ) -> None:
        self._file = file
        self._byteorder = _deduce_byteorder(self._file, byteorder=byteorder)
        self._path = path

    @_annotate_read_exception("logical")
    def read_logical(self) -> bool:
        buf = self._file.read(1)
        return buf != b"\x00"

    @_annotate_read_exception("u8")
    def read_u8(self) -> int:
        buf = self._file.read(1)
        return int.from_bytes(buf, self._byteorder.get())

    @_annotate_read_exception("u32")
    def read_u32(self) -> int:
        buf = self._file.read(4)
        return int.from_bytes(buf, self._byteorder.get())

    @_annotate_read_exception("u32")
    def read_u64(self) -> int:
        buf = self._file.read(8)
        return int.from_bytes(buf, self._byteorder.get())

    @_annotate_read_exception("f64")
    def read_f64(self) -> float:
        buf = self._file.read(8)
        match self._byteorder:
            case Byteorder.little:
                bo = "<"
            case Byteorder.big:
                bo = ">"
        return struct.unpack(bo + "d", buf)[0]  # type: ignore[no-any-return]

    @_annotate_read_exception("char array")
    def read_char_array(self) -> str:
        size = self.read_u32()
        return self.read_n_chars(size)

    @_annotate_read_exception("n chars")
    def read_n_chars(self, n: int) -> str:
        return self._file.read(n).decode("utf-8")

    @_annotate_read_exception("array")
    def read_array(
        self, shape: tuple[int, ...], dtype: np.dtype[_E]
    ) -> npt.NDArray[_E]:
        if not shape:
            return np.array([], dtype=dtype)

        count = int(np.prod(shape))
        dtype = dtype.newbyteorder(self.byteorder.value)
        if isinstance(self._file, BytesIO):
            flat = np.frombuffer(
                self._file.getbuffer(), offset=self.position, dtype=dtype, count=count
            )
            # Make a copy because np.frombuffer creates a view of the buffer.
            # This makes it impossible to write to the buffer while the view exists.
            # But when writing pixel data, we need to both
            # read and write from / to the same buffer.
            flat = flat.copy()
            self._file.seek(self.position + count * dtype.itemsize)
        else:
            flat = np.fromfile(self._file, dtype=dtype, count=int(np.prod(shape)))
        # Invert the shape because files use column-major layout.
        return flat.reshape(shape[::-1])

    @_annotate_write_exception("logical")
    def write_logical(self, value: bool) -> None:
        self._file.write(value.to_bytes(1, self._byteorder.get()))

    @_annotate_write_exception("u8")
    def write_u8(self, value: int) -> None:
        self._file.write(value.to_bytes(1, self._byteorder.get()))

    @_annotate_write_exception("u32")
    def write_u32(self, value: int) -> None:
        self._file.write(value.to_bytes(4, self._byteorder.get()))

    @_annotate_write_exception("u64")
    def write_u64(self, value: int) -> None:
        self._file.write(value.to_bytes(8, self._byteorder.get()))

    @_annotate_write_exception("f64")
    def write_f64(self, value: float) -> None:
        match self._byteorder:
            case Byteorder.little:
                bo = "<"
            case Byteorder.big:
                bo = ">"
        self._file.write(struct.pack(bo + "d", value))

    @_annotate_write_exception("char array")
    def write_char_array(self, value: str) -> None:
        encoded = value.encode("utf-8")
        self.write_u32(len(encoded))
        self._file.write(encoded)

    @_annotate_write_exception("char array")
    def write_chars(self, value: str) -> None:
        encoded = value.encode("utf-8")
        self._file.write(encoded)

    @_annotate_write_exception("array")
    def write_array(
        self, array: npt.NDArray[np.float64] | npt.NDArray[np.float32]
    ) -> None:
        out = array.astype(array.dtype.newbyteorder(self.byteorder.value), copy=False)
        if isinstance(self._file, BytesIO):
            # Inefficient because it constructs an entire separate buffer in memory.
            # Could be optimised to write in chunks if need be.
            self._file.write(out.tobytes())
        else:
            out.tofile(self._file)

    @_annotate_write_exception("bytes")
    def write_raw(self, value: bytes | memoryview) -> None:
        self._file.write(value)

    def seek(self, pos: int) -> None:
        self._file.seek(pos)

    @property
    def byteorder(self) -> Byteorder:
        return self._byteorder

    @property
    def position(self) -> int:
        return self._file.tell()

    @property
    def path(self) -> Path | None:
        return self._path


def _deduce_byteorder(
    file: BinaryIO, *, byteorder: Byteorder | None = None
) -> Byteorder:
    """Guess the byte order of a file.

    The first four bytes of an SQW file are the length of the program name.
    Realistic lengths should be less than 2^16 bytes which is the flip over point
    between little and big endian.
    So we simply use the smaller number.

    This could be made more robust by reading (parts of) the rest of the header
    and checking that it makes sense, e.g., that the program name is valid UTF-8.
    But since HORACE ignores the issue of byteorder, what we have here should be enough.
    """
    if byteorder is not None:
        return byteorder

    pos = file.tell()
    buf = file.read(4)
    file.seek(pos)

    le_size = int.from_bytes(buf, "little")
    be_size = int.from_bytes(buf, "big")
    if le_size < be_size:
        return Byteorder.little
    return Byteorder.big



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

"""SQW fie writer and reader.

This module provides functionality for writing SQW files for
`Horace <https://pace-neutrons.github.io/Horace>`_ and, to a limited extend,
reading SQW files.
The main entrypoint is :class:`Sqw`.

.. rubric:: Classes

.. autosummary::
  :toctree: ../classes
  :template: class-template.rst

  Byteorder
  EnergyMode
  Sqw
  SqwBuilder
  SqwFileType

.. rubric:: Models

.. autosummary::
  :toctree: ../classes
  :template: class-template.rst

  SqwFileHeader
  SqwMainHeader
  SqwIXExperiment
  SqwDndMetadata
  SqwLineAxes
  SqwLineProj
  SqwIXNullInstrument
  SqwIXSample
  SqwIXSource
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

from ._build import SqwBuilder
from ._bytes import Byteorder
from ._models import (
    EnergyMode,
    SqwDndMetadata,
    SqwFileHeader,
    SqwFileType,
    SqwIXExperiment,
    SqwIXNullInstrument,
    SqwIXSample,
    SqwIXSource,
    SqwLineAxes,
    SqwLineProj,
    SqwMainHeader,
)
from ._sqw import Sqw

__all__ = [
    "Byteorder",
    "EnergyMode",
    "Sqw",
    "SqwBuilder",
    "SqwDndMetadata",
    "SqwFileHeader",
    "SqwFileType",
    "SqwIXExperiment",
    "SqwIXNullInstrument",
    "SqwIXSample",
    "SqwIXSource",
    "SqwLineAxes",
    "SqwLineProj",
    "SqwMainHeader",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import enum
from dataclasses import dataclass, field, replace
from datetime import datetime, timezone
from typing import ClassVar, TypeAlias

import numpy as np
import scipp as sc

from . import _ir as ir

DataBlockName: TypeAlias = tuple[str, str]


class SqwFileType(enum.Enum):
    DND = 0
    SQW = 1


@dataclass(frozen=True, kw_only=True, slots=True)
class SqwFileHeader:
    prog_name: str
    prog_version: float
    sqw_type: SqwFileType
    n_dims: int


class SqwDataBlockType(enum.Enum):
    regular = "data_block"
    pix = "pix_data_block"
    dnd = "dnd_data_block"


@dataclass(frozen=True, kw_only=True, slots=True)
class SqwDataBlockDescriptor:
    block_type: SqwDataBlockType
    name: DataBlockName
    position: int  # u64
    size: int  # u32
    locked: bool  # u32


@dataclass(kw_only=True, slots=True)
class SqwMainHeader(ir.Serializable):
    full_filename: str
    title: str
    nfiles: int
    creation_date: datetime

    serial_name: ClassVar[str] = "main_header_cl"
    version: ClassVar[float] = 2.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "full_filename": ir.String(self.full_filename),
            "title": ir.String(self.title),
            "nfiles": ir.F64(float(self.nfiles)),
            "creation_date": ir.Datetime(self.creation_date),
            "creation_date_defined_privately": ir.Logical(False),
        }

    def prepare_for_serialization(self, filename: str, filepath: str) -> SqwMainHeader:
        return replace(self, creation_date=datetime.now(tz=timezone.utc))


@dataclass(kw_only=True, slots=True)
class SqwLineAxes(ir.Serializable):
    title: str
    label: list[str]
    img_scales: list[sc.Variable]
    img_range: list[sc.Variable]
    n_bins_all_dims: sc.Variable  # shape=(n_dim,) dtype=float64 [encodes int]
    single_bin_defines_iax: sc.Variable  # shape=(n_dim,) dtype=bool
    dax: sc.Variable
    offset: list[sc.Variable]
    changes_aspect_ratio: bool
    filename: str = ""
    filepath: str = ""

    serial_name: ClassVar[str] = "line_axes"
    version: ClassVar[float] = 7.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        units = ["1/angstrom"] * 3 + ["meV"]  # depends on SqwLineProj.type

        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "filename": ir.String(self.filename),
            "filepath": ir.String(self.filepath),
            "title": ir.String(self.title),
            "label": _serialize_str_array(self.label),
            "img_scales": _serialize_multi_unit_array(self.img_scales, units),
            "img_range": _serialize_multi_unit_array(self.img_range, units),
            "nbins_all_dims": _variable_to_float_array(self.n_bins_all_dims, None),
            "single_bin_defines_iax": ir.ObjectArray(
                shape=self.n_bins_all_dims.shape,
                data=[ir.Logical(bool(b)) for b in self.single_bin_defines_iax.values],
                ty=ir.TypeTag.logical,
            ),
            # +1 to convert to 1-based indexing
            "dax": _variable_to_float_array(self.dax + sc.index(1), None),
            "offset": _serialize_multi_unit_array(self.offset, units),
            "changes_aspect_ratio": ir.Logical(self.changes_aspect_ratio),
        }


@dataclass(kw_only=True, slots=True)
class SqwLineProj(ir.Serializable):
    lattice_spacing: sc.Variable  # vector
    lattice_angle: sc.Variable  # vector
    offset: list[sc.Variable]
    title: str
    label: list[str]
    u: sc.Variable  # vector
    v: sc.Variable  # vector
    w: sc.Variable | None  # vector
    non_orthogonal: bool
    type: str

    serial_name: ClassVar[str] = "line_proj"
    version: ClassVar[float] = 7.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        if self.type != "aaa":
            raise NotImplementedError(f"Projection type not supported: {self.type}")
        units = ["1/angstrom"] * 3 + ["meV"]  # depends on SqwLineProj.type

        if self.w is None:
            w = ir.Array(np.array([]), ir.TypeTag.f64)
        else:
            w = _variable_to_float_array(self.w, "1/angstrom")

        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "alatt": _variable_to_float_array(self.lattice_spacing, "angstrom"),
            "angdeg": _variable_to_float_array(self.lattice_angle, "deg"),
            "offset": _serialize_multi_unit_array(self.offset, units),
            "title": ir.String(self.title),
            "label": _serialize_str_array(self.label),
            "u": _variable_to_float_array(self.u, "1/angstrom"),
            "v": _variable_to_float_array(self.v, "1/angstrom"),
            "w": w,
            "nonorthogonal": ir.Logical(self.non_orthogonal),
            "type": ir.String(self.type),
        }


@dataclass(kw_only=True, slots=True)
class SqwDndMetadata(ir.Serializable):
    axes: SqwLineAxes
    proj: SqwLineProj
    creation_date: datetime = field(
        default_factory=lambda: datetime.now(tz=timezone.utc)
    )

    serial_name: ClassVar[str] = "dnd_metadata"
    version: ClassVar[float] = 1.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        axes = self.axes.serialize_to_ir()
        proj = self.proj.serialize_to_ir()

        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "axes": ir.ObjectArray(
                ty=ir.TypeTag.struct,
                shape=(1,),
                data=[axes],
            ),
            "proj": ir.ObjectArray(
                ty=ir.TypeTag.struct,
                shape=(1,),
                data=[proj],
            ),
            "creation_date_str": ir.Datetime(self.creation_date),
        }

    def prepare_for_serialization(self, filename: str, filepath: str) -> SqwDndMetadata:
        return replace(
            self,
            creation_date=datetime.now(tz=timezone.utc),
            axes=replace(self.axes, filename=filename, filepath=filepath),
        )


@dataclass(kw_only=True, slots=True)
class SqwPixelMetadata(ir.Serializable):
    full_filename: str
    npix: int
    data_range: np.ndarray[tuple[int, int], np.dtype[np.float64]]

    serial_name: ClassVar[str] = "pix_metadata"
    version: ClassVar[float] = 1.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "full_filename": ir.String(self.full_filename),
            "npix": ir.F64(float(self.npix)),
            "data_range": ir.Array(self.data_range, ty=ir.TypeTag.f64),
        }


@dataclass(kw_only=True, slots=True)
class SqwPixWrap(ir.Serializable):
    """Represents pixel data but does not hold the actual data."""

    n_rows: int = 9
    n_pixels: int

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "n_rows": ir.U32(self.n_rows),
            "n_pixels": ir.U64(self.n_pixels),
        }


@dataclass(kw_only=True, slots=True)
class SqwIXSource(ir.Serializable):
    name: str
    target_name: str
    frequency: sc.Variable

    serial_name: ClassVar[str] = "IX_source"
    version: ClassVar[float] = 2.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "name": ir.String(self.name),
            "target_name": ir.String(self.target_name),
            "frequency": ir.F64(self.frequency.value),  # TODO unit
        }


@dataclass(kw_only=True, slots=True)
class SqwIXNullInstrument(ir.Serializable):
    name: str
    source: SqwIXSource

    serial_name: ClassVar[str] = "IX_null_inst"
    version: ClassVar[float] = 2.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "source": self.source.serialize_to_ir(),
            "name": ir.String(self.name),
        }


class EnergyMode(enum.Enum):
    direct = 1
    indirect = 2


@dataclass(kw_only=True, slots=True)
class SqwIXSample(ir.Serializable):
    name: str
    lattice_spacing: sc.Variable  # vector
    lattice_angle: sc.Variable  # vector

    serial_name: ClassVar[str] = "IX_sample"
    version: ClassVar[float] = 3.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "alatt": _variable_to_float_array(self.lattice_spacing, "angstrom"),
            "angdeg": _variable_to_float_array(self.lattice_angle, "deg"),
            "name": ir.String(self.name),
        }


# In contrast to SQW files, this model contains the nested
# struct fields instead of a nested struct in `array-dat`.
@dataclass(kw_only=True, slots=True)
class SqwIXExperiment(ir.Serializable):
    run_id: int
    # 1 element for direct, array of detector.shape for indirect
    efix: sc.Variable  # array or scalar
    emode: EnergyMode
    en: sc.Variable  # array
    psi: sc.Variable  # scalar
    u: sc.Variable  # vector
    v: sc.Variable  # vector
    omega: sc.Variable  # scalar
    dpsi: sc.Variable  # scalar
    gl: sc.Variable  # scalar
    gs: sc.Variable  # scalar
    filename: str = ""
    filepath: str = ""

    serial_name: ClassVar[str] = "IX_experiment"
    version: ClassVar[float] = 3.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        en = self.en.to(unit="meV", dtype="float64", copy=False)
        if en.ndim == 1:  # direct mode; still needs 2D en like indirect mode
            en = en.broadcast(sizes={"_": 1, "energy_transfer": self.en.shape[0]})
        else:
            en = en.transpose(dims=["detector", "energy_transfer"])

        efix = self.efix.to(unit="meV", dtype="float64", copy=False)
        if efix.ndim == 0:
            efix = efix.broadcast(sizes={"_": 1})
        return {
            "filename": ir.String(self.filename),
            "filepath": ir.String(self.filepath),
            "run_id": ir.F64(float(self.run_id + 1)),
            "efix": ir.Array(efix.values, ty=ir.TypeTag.f64),
            "emode": ir.F64(float(self.emode.value)),
            "en": ir.Array(en.values, ty=ir.TypeTag.f64),
            "psi": ir.F64(_angle_value(self.psi)),
            "u": ir.Array(self.u.values, ty=ir.TypeTag.f64),
            "v": ir.Array(self.v.values, ty=ir.TypeTag.f64),
            "omega": ir.F64(_angle_value(self.omega)),
            "dpsi": ir.F64(_angle_value(self.dpsi)),
            "gl": ir.F64(_angle_value(self.gl)),
            "gs": ir.F64(_angle_value(self.gs)),
            "angular_is_degree": ir.Logical(False),
            # serial_name and version are serialized by SqwMultiIXExperiment
        }


@dataclass(slots=True)
class SqwMultiIXExperiment(ir.Serializable):
    array_dat: list[SqwIXExperiment]

    serial_name: ClassVar[str] = "IX_experiment"
    version: ClassVar[float] = 3.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "array_dat": ir.ObjectArray(
                ty=ir.TypeTag.struct,
                shape=(len(self.array_dat),),
                data=[exp.serialize_to_ir() for exp in self.array_dat],
            ),
        }


@dataclass(kw_only=True, slots=True)
class UniqueRefContainer(ir.Serializable):
    global_name: str
    objects: UniqueObjContainer

    serial_name: ClassVar[str] = "unique_references_container"
    version: ClassVar[float] = 1.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "stored_baseclass": ir.String(self.objects.baseclass),
            "global_name": ir.String(self.global_name),
            "unique_objects": self.objects.serialize_to_ir().to_object_array(),
        }


@dataclass(kw_only=True, slots=True)
class UniqueObjContainer(ir.Serializable):
    baseclass: str
    objects: list[ir.Serializable]
    indices: list[int]

    serial_name: ClassVar[str] = "unique_objects_container"
    version: ClassVar[float] = 1.0

    def _serialize_to_dict(
        self,
    ) -> dict[str, ir.Object | ir.ObjectArray | ir.CellArray]:
        return {
            "serial_name": ir.String(self.serial_name),
            "version": ir.F64(self.version),
            "baseclass": ir.String(self.baseclass),
            "unique_objects": ir.CellArray(
                shape=(len(self.objects),),
                data=[obj.serialize_to_ir().to_object_array() for obj in self.objects],
            ),
            "idx": ir.ObjectArray(
                shape=(len(self.indices),),
                # +1 to convert to 1-based indexing
                data=np.array(self.indices) + 1.0,
                ty=ir.TypeTag.f64,
            ),
        }


def _angle_value(x: sc.Variable) -> float:
    return x.to(unit="rad", dtype="float64", copy=False).value  # type: ignore[no-any-return]


def _serialize_str_array(strings: list[str]) -> ir.CellArray:
    return ir.CellArray(
        shape=(len(strings),),
        data=[
            ir.ObjectArray(shape=(len(s),), data=[ir.String(s)], ty=ir.TypeTag.char)
            for s in strings
        ],
    )


def _serialize_multi_unit_array(data: list[sc.Variable], units: list[str]) -> ir.Array:
    stacked = np.stack(
        [d.to(unit=u, dtype="float64").values for d, u in zip(data, units, strict=True)]
    )
    return ir.Array(stacked, ty=ir.TypeTag.f64)


def _variable_to_float_array(var: sc.Variable, unit: str | None) -> ir.Array:
    if unit is not None:
        var = var.to(unit=unit, copy=False)
    elif var.unit is not None:
        raise sc.UnitError(f"Expected no unit, got: {var.unit}")
    return ir.Array(
        var.values.astype("float64", copy=False),
        ir.TypeTag.f64,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""Metadata utilities.

.. rubric:: Models

.. autosummary::
  :toctree: ../classes
  :template: model-template.rst

  Beamline
  Measurement
  Person
  Software
  Source

.. rubric:: Auxiliary Classes

.. autosummary::
  :toctree: ../classes
  :template: class-template.rst

  ORCIDiD
  RadiationProbe
  SourceType

.. rubric:: Attributes

.. autosummary::

  ESS_SOURCE
"""

from ._model import (
    ESS_SOURCE,
    Beamline,
    Measurement,
    Person,
    RadiationProbe,
    Software,
    Source,
    SourceType,
)
from ._orcid import ORCIDiD

__all__ = [
    'ESS_SOURCE',
    'Beamline',
    'Measurement',
    'ORCIDiD',
    'Person',
    'RadiationProbe',
    'Software',
    'Source',
    'SourceType',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from datetime import datetime
from io import BytesIO
from pathlib import Path

import pytest
import scipp as sc
import scipp.testing

from scippneutron.io.sqw import Byteorder, EnergyMode, Sqw, SqwFileHeader, SqwFileType


@pytest.fixture(params=["file", "buffer"])
def intact_v4_sqw(request: pytest.FixtureRequest) -> Path | BytesIO:
    from scippneutron.data import get_path

    path = Path(get_path("horace_sqw_4d.sqw"))
    if request.param == "buffer":
        return BytesIO(path.read_bytes())
    return path


def test_detects_byteorder_little_endian() -> None:
    buf = BytesIO(
        b"\x06\x00\x00\x00"
        b"horace"
        b"\x00\x00\x00\x00\x00\x00\x10\x40"
        b"\x01\x00\x00\x00"
        b"\x04\x00\x00\x00"
    )
    with Sqw.open(buf) as sqw:
        assert sqw.byteorder == Byteorder.little


def test_detects_byteorder_big_endian() -> None:
    buf = BytesIO(
        b"\x00\x00\x00\x06"
        b"horace"
        b"\x40\x10\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x01"
        b"\x00\x00\x00\x04"
    )
    with Sqw.open(buf) as sqw:
        assert sqw.byteorder == Byteorder.big


def test_open_file_header_little_endian() -> None:
    buf = BytesIO(
        b"\x06\x00\x00\x00"
        b"horace"
        b"\x00\x00\x00\x00\x00\x00\x10\x40"
        b"\x01\x00\x00\x00"
        b"\x04\x00\x00\x00"
    )
    expected = SqwFileHeader(
        prog_name="horace",
        prog_version=4.0,
        sqw_type=SqwFileType.SQW,
        n_dims=4,
    )
    with Sqw.open(buf) as sqw:
        assert sqw.file_header == expected


def test_open_file_header_big_endian() -> None:
    buf = BytesIO(
        b"\x00\x00\x00\x06"
        b"horace"
        b"\x40\x10\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x01"
        b"\x00\x00\x00\x04"
    )
    expected = SqwFileHeader(
        prog_name="horace",
        prog_version=4.0,
        sqw_type=SqwFileType.SQW,
        n_dims=4,
    )
    with Sqw.open(buf) as sqw:
        assert sqw.file_header == expected


def test_open_flags_wrong_prog_name() -> None:
    buf = BytesIO(
        b"\x07\x00\x00\x00"
        b"sqomega"
        b"\x00\x00\x00\x00\x00\x00\x10\x40"
        b"\x01\x00\x00\x00"
        b"\x04\x00\x00\x00"
    )
    expected = SqwFileHeader(
        prog_name="sqomega",
        prog_version=4.0,
        sqw_type=SqwFileType.SQW,
        n_dims=4,
    )
    with pytest.warns(UserWarning, match="SQW program not supported"):
        with Sqw.open(buf) as sqw:
            assert sqw.file_header == expected


def test_open_flags_wrong_prog_version() -> None:
    buf = BytesIO(
        b"\x06\x00\x00\x00"
        b"horace"
        b"\x00\x00\x00\x00\x00\x00\x20\x40"
        b"\x01\x00\x00\x00"
        b"\x04\x00\x00\x00"
    )
    expected = SqwFileHeader(
        prog_name="horace",
        prog_version=8.0,
        sqw_type=SqwFileType.SQW,
        n_dims=4,
    )
    with pytest.warns(UserWarning, match="SQW program not supported"):
        with Sqw.open(buf) as sqw:
            assert sqw.file_header == expected


def test_read_data_block_raises_when_given_tuple_and_str(
    intact_v4_sqw: Path | BytesIO,
) -> None:
    with Sqw.open(intact_v4_sqw) as sqw:
        with pytest.raises(TypeError):
            sqw.read_data_block(("", "main_header"), "extra")


def test_read_data_block_raises_when_given_only_one_str(
    intact_v4_sqw: Path | BytesIO,
) -> None:
    with Sqw.open(intact_v4_sqw) as sqw:
        with pytest.raises(TypeError):
            sqw.read_data_block("main_header")


def test_read_main_header(intact_v4_sqw: Path | BytesIO) -> None:
    with Sqw.open(intact_v4_sqw) as sqw:
        main_header = sqw.read_data_block(("", "main_header"))
    assert main_header.version == 2.0
    assert main_header.title == ""
    assert type(main_header.nfiles) is int  # because it is encoded as f64 in file
    assert main_header.nfiles == 23
    # TODO can we encode a timezone? How does horace react?
    assert main_header.creation_date == datetime(2024, 3, 21, 21, 16, 56)  # noqa: DTZ001


def test_read_expdata(intact_v4_sqw: Path | BytesIO) -> None:
    with Sqw.open(intact_v4_sqw) as sqw:
        main_header = sqw.read_data_block("", "main_header")
        expdata = sqw.read_data_block("experiment_info", "expdata")
    assert len(expdata) == main_header.nfiles
    assert expdata[0].emode == EnergyMode.direct
    # Numbers copied from HORACE:
    sc.testing.assert_identical(
        expdata[0].psi, sc.scalar(-0.03490658476948738, unit='rad')
    )
    sc.testing.assert_identical(
        expdata[0].u,
        sc.vector([0.9957987070083618, 0.002324522938579321, -0.004358167294412851]),
    )


def test_read_sample(intact_v4_sqw: Path | BytesIO) -> None:
    with Sqw.open(intact_v4_sqw) as sqw:
        main_header = sqw.read_data_block("", "main_header")
        samples = sqw.read_data_block("experiment_info", "samples")
    assert len(samples) == main_header.nfiles
    assert all(sample == samples[0] for sample in samples[1:])
    sc.testing.assert_identical(
        samples[0].lattice_spacing,
        sc.vector([2.8579773902893066] * 3, unit="1/angstrom"),
    )
    sc.testing.assert_identical(
        samples[0].lattice_angle, sc.vector([90, 90, 90], unit="deg")
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

from collections.abc import Generator

import pytest
import scipp as sc

from scippneutron.conversion.graph import beamline, tof


def _flatten_keys(graph: dict[str, object]) -> Generator[str | tuple[str], None, None]:
    for key in graph:
        if isinstance(key, str):
            yield key
        else:
            yield from key


@pytest.mark.parametrize("target", _flatten_keys(tof.elastic('tof')))
def test_elastic_can_compute_all_from_tof(target: str | tuple[str]) -> None:
    graph = {**tof.elastic('tof'), **beamline.beamline(scatter=True)}
    da = sc.DataArray(
        sc.zeros(sizes={'tof': 4, 'detector': 2}),
        coords={
            'tof': sc.arange('tof', 5, unit='us'),
            'pulse_time': 0.1 * sc.arange('tof', 5, unit='us'),
            'position': sc.vectors(
                dims=['detector'], values=[[1, 2, 3], [4, 2, 1]], unit='m'
            ),
            'sample_position': sc.vector(value=[0, 0, 0], unit='m'),
            'source_position': sc.vector(value=[0, 0, -10], unit='m'),
            'sample_rotation': sc.spatial.rotation(value=[0, 0, 1, 1]),
            'b_matrix': sc.spatial.linear_transform(
                value=[[1, 0, 0], [0, 1, 0], [0, 1, 1]]
            ),
            'u_matrix': sc.spatial.linear_transform(
                value=[[0, 0.5, 0], [0.2, 1, 0], [0.1, 0, 1]]
            ),
        },
    )
    result = da.transform_coords(target, graph)
    assert target in result.coords



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import warnings
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Any, BinaryIO, Literal

import numpy as np
import numpy.typing as npt
import scipp as sc
from dateutil.parser import parse as parse_datetime

from .._files import open_or_pass
from . import _ir as ir
from ._build import SqwBuilder
from ._bytes import Byteorder
from ._low_level_io import LowLevelSqw
from ._models import (
    DataBlockName,
    EnergyMode,
    SqwDataBlockDescriptor,
    SqwDataBlockType,
    SqwDndMetadata,
    SqwFileHeader,
    SqwFileType,
    SqwIXExperiment,
    SqwIXNullInstrument,
    SqwIXSample,
    SqwIXSource,
    SqwLineAxes,
    SqwLineProj,
    SqwMainHeader,
    SqwPixelMetadata,
)
from ._read_write import read_object_array


class Sqw:
    def __init__(
        self,
        *,
        sqw_io: LowLevelSqw,
        file_header: SqwFileHeader,
        block_allocation_table: dict[DataBlockName, SqwDataBlockDescriptor],
    ) -> None:
        self._sqw_io = sqw_io
        self._file_header = file_header
        self._block_allocation_table = block_allocation_table

    @classmethod
    @contextmanager
    def open(
        cls,
        path: str | PathLike[str] | BinaryIO | BytesIO,
        *,
        byteorder: Byteorder | Literal["little", "big"] | None = None,
    ) -> Generator[Sqw, None, None]:
        with open_or_pass(path, "rb") as f:
            stored_path = None if isinstance(path, BinaryIO | BytesIO) else Path(path)
            sqw_io = LowLevelSqw(
                f,
                path=stored_path,
                byteorder=Byteorder.parse(byteorder) if byteorder is not None else None,
            )
            file_header = _read_file_header(sqw_io)
            data_block_descriptors = _read_block_allocation_table(sqw_io)
            yield Sqw(
                sqw_io=sqw_io,
                file_header=file_header,
                block_allocation_table=data_block_descriptors,
            )

    @classmethod
    def build(
        cls,
        path: str | PathLike[str] | BinaryIO | BytesIO,
        *,
        title: str = "",
        byteorder: Byteorder | Literal["native", "little", "big"] = "native",
    ) -> SqwBuilder:
        return SqwBuilder(path, title, byteorder=Byteorder.parse(byteorder))

    @property
    def file_header(self) -> SqwFileHeader:
        return self._file_header

    @property
    def byteorder(self) -> Byteorder:
        return self._sqw_io.byteorder

    def data_block_names(self) -> Iterable[DataBlockName]:
        """Return an iterator over the names of all stored data blocks."""
        return self._block_allocation_table.keys()

    # TODO lock blocks during writing, esp pix data
    def read_data_block(
        self, name: DataBlockName | str, level2_name: str | None = None, /
    ) -> Any:  # TODO type
        block_name = _normalize_data_block_name(name, level2_name)
        try:
            block_descriptor = self._block_allocation_table[block_name]
        except KeyError:
            raise KeyError(f"No data block {block_name!r} in file") from None

        self._sqw_io.seek(block_descriptor.position)
        match block_descriptor.block_type:
            case SqwDataBlockType.regular:
                return _parse_block(read_object_array(self._sqw_io))
            case SqwDataBlockType.pix:
                return _read_pix_block(self._sqw_io)
            case SqwDataBlockType.dnd:
                return _read_dnd_block(self._sqw_io)
            case _:
                raise NotImplementedError(
                    f"Unsupported data block type: {block_descriptor.block_type}"
                )

    def __str__(self) -> str:
        path = self._sqw_io.path
        path_piece = "In-memory SQW file" if path is None else f"SQW file '{path}'"
        program = self._file_header.prog_name
        version = self._file_header.prog_version
        sqw_type = self._file_header.sqw_type
        n_dims = self._file_header.n_dims
        return f"{path_piece} ({sqw_type}, {program=}, {version=}, {n_dims=})"


def _read_file_header(sqw_io: LowLevelSqw) -> SqwFileHeader:
    prog_name = sqw_io.read_char_array()
    prog_version = sqw_io.read_f64()
    if prog_name != "horace" or prog_version != 4.0:
        warnings.warn(
            f"SQW program not supported: '{prog_name}' version {prog_version} "
            f"(expected 'horace' with version 4.0)",
            UserWarning,
            stacklevel=2,
        )

    sqw_type = SqwFileType(sqw_io.read_u32())
    if sqw_type != SqwFileType.SQW:
        warnings.warn("DND files are not supported", UserWarning, stacklevel=2)

    n_dims = sqw_io.read_u32()
    return SqwFileHeader(
        prog_name=prog_name, prog_version=prog_version, sqw_type=sqw_type, n_dims=n_dims
    )


def _read_block_allocation_table(
    sqw_io: LowLevelSqw,
) -> dict[DataBlockName, SqwDataBlockDescriptor]:
    _bat_size = sqw_io.read_u32()  # don't need this
    n_blocks = sqw_io.read_u32()
    descriptors = [_read_data_block_descriptor(sqw_io) for _ in range(n_blocks)]
    return {descriptor.name: descriptor for descriptor in descriptors}


def _read_data_block_descriptor(sqw_io: LowLevelSqw) -> SqwDataBlockDescriptor:
    block_type = SqwDataBlockType(sqw_io.read_char_array())
    name = sqw_io.read_char_array(), sqw_io.read_char_array()
    position = sqw_io.read_u64()
    size = sqw_io.read_u32()
    locked = sqw_io.read_u32() == 1
    return SqwDataBlockDescriptor(
        block_type=block_type, name=name, position=position, size=size, locked=locked
    )


def _normalize_data_block_name(
    name: DataBlockName | str, level2_name: str | None
) -> DataBlockName:
    match (name, level2_name):
        case (str(n1), str(n2)), None:
            return DataBlockName((n1, n2))  # type: ignore[no-any-return, operator]
        case str(n1), str(n2):
            return DataBlockName((n1, n2))  # type: ignore[no-any-return, operator]
    raise TypeError(
        "Data block name must be given either as a tuple of two strings or two "
        f"separate strings. Got {name!r} and {level2_name!r}."
    )


class AbortParse(Exception): ...


def _parse_block(block: ir.ObjectArray | ir.CellArray) -> Any:
    try:
        if isinstance(block, ir.CellArray):
            raise AbortParse("Block is a cell array, not a struct")
        return _try_parse_block(block)
    except AbortParse as abort:
        warnings.warn(
            f"Unable to parse SQW block: {abort.args[0]}", UserWarning, stacklevel=2
        )
        return block


def _try_parse_block(block: ir.ObjectArray) -> Any:
    if block.shape != (1,):
        raise AbortParse(f"Unsupported block shape: {block.shape}")
    if block.ty != ir.TypeTag.struct:
        raise AbortParse(f"Unsupported block type: {block.shape}, expected struct")
    struct: ir.Struct = block.data[0]  # type: ignore[assignment]
    if sum(s != 1 for s in struct.field_values.shape) != 1:
        raise AbortParse(
            "Contents cannot be a multi-dimensional cell array, "
            f"got shape {struct.field_values.shape}"
        )

    key = _get_struct_type_id(struct)
    try:
        parser = _BLOCK_PARSERS[key]
    except KeyError:
        raise AbortParse(f"No parser for struct type {key}") from None
    return parser(struct)


def _get_struct_type_id(struct: ir.Struct) -> tuple[str, float]:
    name = _get_struct_field(struct, "serial_name")
    if len(name.shape) != 1:
        raise AbortParse("'serial_name' is multi-dimensional")
    version = _get_struct_field(struct, "version")
    if version.shape != (1,):
        raise AbortParse("'version' is multi-dimensional")

    n = name.data[0]
    v = version.data[0]
    if not isinstance(n, ir.String):
        raise AbortParse("'serial_name' is not a string")
    if not isinstance(v, ir.F64):
        raise AbortParse("'version' is not an f64")

    return n.value, v.value


def _get_struct_field(struct: ir.Struct, name: str) -> ir.ObjectArray | ir.CellArray:
    for field_name, value in zip(
        struct.field_names, struct.field_values.data, strict=True
    ):
        if field_name == name:
            return value
    raise AbortParse(f"No field '{name}' in struct")


def _get_scalar_struct_field(struct: ir.Struct, name: str) -> Any:
    field = _get_struct_field(struct, name)
    shape = field.shape[1:] if field.ty == ir.TypeTag.char else field.shape
    if shape not in ((1,), ()):
        raise AbortParse(f"Field '{name}' has non-scalar shape: {shape}")
    if isinstance(field.data[0], ir.Struct | ir.ObjectArray | ir.CellArray):
        raise AbortParse(f"Field '{name}' contains a nested struct or cell array")
    return field.data[0].value


def _unpack_cell_array(cell_array: ir.CellArray) -> list[Any]:
    # This does not support general cell arrays.
    # It is specialized for 1d cell arrays of strings.
    data = (obj.data for obj in cell_array.data)
    return [d[0].value if len(d) == 1 else [x.value for x in d] for d in data]  # type: ignore[union-attr]


def _parse_main_header_cl_2_0(struct: ir.Struct) -> SqwMainHeader:
    return SqwMainHeader(
        full_filename=_get_scalar_struct_field(struct, "full_filename"),
        title=_get_scalar_struct_field(struct, "title"),
        nfiles=int(_get_scalar_struct_field(struct, "nfiles")),
        creation_date=parse_datetime(_get_scalar_struct_field(struct, "creation_date")),
    )


def _parse_dnd_metadata_1_0(struct: ir.Struct) -> SqwDndMetadata:
    (axes_struct,) = _get_struct_field(struct, "axes").data
    (proj_struct,) = _get_struct_field(struct, "proj").data
    proj, units = _parse_line_proj_7_0(proj_struct)  # type: ignore[arg-type]
    axes = _parse_line_axes_7_0(axes_struct, units)  # type: ignore[arg-type]
    return SqwDndMetadata(
        axes=axes,
        proj=proj,
        creation_date=parse_datetime(
            _get_scalar_struct_field(struct, "creation_date_str")
        ),
    )


def _parse_line_axes_7_0(struct: ir.Struct, units: list[str]) -> SqwLineAxes:
    return SqwLineAxes(
        title=_get_scalar_struct_field(struct, "title"),
        label=_unpack_cell_array(_get_struct_field(struct, "label")),
        img_scales=_parse_1d_multi_unit_array(
            _get_struct_field(struct, "img_scales").data, units
        ),
        img_range=_parse_2d_multi_unit_array(
            "range", _get_struct_field(struct, "img_range").data, units
        ),
        n_bins_all_dims=sc.array(
            dims=("axis",),
            values=_get_struct_field(struct, "nbins_all_dims").data,
            dtype="int64",
            unit=None,
        ),
        single_bin_defines_iax=sc.array(
            dims=("axis",),
            values=[
                d.value
                for d in _get_struct_field(struct, "single_bin_defines_iax").data
            ],
        ),
        dax=sc.array(
            dims=("axis",),
            # -1 to convert to 0-based indexing
            values=_get_struct_field(struct, "dax").data.astype(int) - 1,
            unit=None,
        ),
        offset=_parse_1d_multi_unit_array(
            _get_struct_field(struct, "offset").data, units
        ),
        changes_aspect_ratio=_get_scalar_struct_field(struct, "changes_aspect_ratio"),
        filename=_get_scalar_struct_field(struct, "filename"),
        filepath=_get_scalar_struct_field(struct, "filepath"),
    )


def _parse_line_proj_7_0(struct: ir.Struct) -> tuple[SqwLineProj, list[str]]:
    def get_vec(name: str, unit: str) -> sc.Variable | None:
        values = _get_struct_field(struct, name).data
        if values.shape == (0,):
            return None
        return sc.vector(values, unit=unit)

    ty = _get_scalar_struct_field(struct, "type")
    if ty != "aaa":
        # See https://pace-neutrons.github.io/Horace/v4.0.0/manual/Cutting_data_of_interest_from_SQW_files_and_objects.html#projection-in-details
        # for other units
        raise AbortParse(f"Unsupported 'type' in line_proj: {ty}")
    units = ["1/angstrom"] * 3 + ["meV"]

    return SqwLineProj(
        lattice_spacing=sc.vector(
            _get_struct_field(struct, "alatt").data, unit="1/angstrom"
        ),
        lattice_angle=sc.vector(_get_struct_field(struct, "angdeg").data, unit="deg"),
        offset=_parse_1d_multi_unit_array(
            _get_struct_field(struct, "offset").data,
            units,
        ),
        title=_get_scalar_struct_field(struct, "title"),
        label=_unpack_cell_array(_get_struct_field(struct, "label")),
        u=get_vec("u", unit=units[0]),
        v=get_vec("v", unit=units[1]),
        w=get_vec("w", unit=units[2]),
        non_orthogonal=_get_scalar_struct_field(struct, "nonorthogonal"),
        type=ty,
    ), units


def _parse_1d_multi_unit_array(
    values: npt.NDArray[Any], units: list[str]
) -> list[sc.Variable]:
    return [
        sc.scalar(value, unit=unit) for value, unit in zip(values, units, strict=False)
    ]


def _parse_2d_multi_unit_array(
    dim: str, values: npt.NDArray[Any], units: list[str]
) -> list[sc.Variable]:
    return [
        sc.array(dims=[dim], values=values, unit=unit)
        for values, unit in zip(values, units, strict=False)
    ]


def _parse_pix_metadata_1_0(struct: ir.Struct) -> SqwPixelMetadata:
    data_range = _get_struct_field(struct, "data_range").data
    if not isinstance(data_range, np.ndarray):
        raise AbortParse("'data_range' is not a numpy array")
    return SqwPixelMetadata(
        full_filename=_get_scalar_struct_field(struct, "full_filename"),
        npix=int(_get_scalar_struct_field(struct, "npix")),
        data_range=data_range,
    )


def _parse_ix_source_2_0(struct: ir.Struct) -> SqwIXSource:
    name = _get_scalar_struct_field(struct, "name")
    target_name = _get_scalar_struct_field(struct, "target_name")
    frequency = _get_scalar_struct_field(struct, "frequency")
    return SqwIXSource(
        name=name,
        target_name=target_name,
        # TODO unit currently unknown
        frequency=sc.scalar(frequency, unit=None),
    )


def _parse_ix_null_instrument_1_0(struct: ir.Struct) -> SqwIXNullInstrument:
    source = _try_parse_block(_get_struct_field(struct, "source"))
    name = _get_scalar_struct_field(struct, "name")
    return SqwIXNullInstrument(name=name, source=source)


def _parse_ix_sample_0_0(struct: ir.Struct) -> SqwIXSample:
    name = _get_scalar_struct_field(struct, "name")
    lattice_spacing = sc.vector(
        _get_struct_field(struct, "alatt").data, unit="1/angstrom"
    )
    lattice_angle = sc.vector(_get_struct_field(struct, "angdeg").data, unit="deg")
    return SqwIXSample(
        name=name,
        lattice_spacing=lattice_spacing,
        lattice_angle=lattice_angle,
    )


def _parse_ix_experiment_3_0(struct: ir.Struct) -> list[SqwIXExperiment]:
    return [
        _parse_single_ix_experiment_3_0(run)
        for run in _get_struct_field(struct, "array_dat").data
    ]


def _parse_single_ix_experiment_3_0(struct: ir.Struct) -> SqwIXExperiment:
    def g(n: str) -> Any:
        return _get_scalar_struct_field(struct, n)

    candidate_efix = _get_struct_field(struct, "efix").data
    if isinstance(candidate_efix, np.ndarray):
        efix = sc.array(dims=["detector"], values=candidate_efix, unit="meV")
    else:
        (e,) = candidate_efix
        efix = sc.scalar(e.value, unit="meV")

    raw_en = _get_struct_field(struct, "en").data
    if isinstance(raw_en, np.ndarray):
        en = raw_en.squeeze()
    else:
        en = [e.value for e in raw_en]

    angle_unit = sc.Unit("deg" if g("angular_is_degree") else "rad")

    return SqwIXExperiment(
        filename=g("filename"),
        filepath=g("filepath"),
        run_id=int(g("run_id")) - 1,
        efix=efix,
        emode=EnergyMode(g("emode")),
        en=sc.array(dims=["energy_transfer"], values=en, unit="meV"),
        psi=sc.scalar(g("psi"), unit=angle_unit),
        u=sc.vector(_get_struct_field(struct, "u").data),
        v=sc.vector(_get_struct_field(struct, "v").data),
        omega=sc.scalar(g("omega"), unit=angle_unit),
        dpsi=sc.scalar(g("dpsi"), unit=angle_unit),
        gl=sc.scalar(g("gl"), unit=angle_unit),
        gs=sc.scalar(g("gs"), unit=angle_unit),
    )


def _parse_unique_references_container_1_0(struct: ir.Struct) -> list[Any]:
    objects = _get_struct_field(struct, "unique_objects").data
    if len(objects) != 1:
        raise AbortParse(
            "Expected exactly one object in a 'unique_references_container', "
            f"got {len(objects)}"
        )
    return _parse_unique_objects_container_1_0(objects[0])


def _parse_unique_objects_container_1_0(struct: ir.Struct) -> list[Any]:
    objects = _get_struct_field(struct, "unique_objects").data
    idx = _get_struct_field(struct, "idx").data
    if not isinstance(idx, np.ndarray):  # it is list[ir.F64]
        idx = np.array([i.value for i in idx])
    parsed_objects = [_try_parse_block(obj) for obj in objects]
    # -1 to convert to 0-based index
    return [parsed_objects[int(i) - 1] for i in idx]


_BLOCK_PARSERS = {
    (SqwMainHeader.serial_name, SqwMainHeader.version): _parse_main_header_cl_2_0,
    (SqwDndMetadata.serial_name, SqwDndMetadata.version): _parse_dnd_metadata_1_0,
    (SqwPixelMetadata.serial_name, SqwPixelMetadata.version): _parse_pix_metadata_1_0,
    (
        SqwIXNullInstrument.serial_name,
        SqwIXNullInstrument.version,
    ): _parse_ix_null_instrument_1_0,
    (SqwIXExperiment.serial_name, SqwIXExperiment.version): _parse_ix_experiment_3_0,
    (SqwIXSample.serial_name, SqwIXSample.version): _parse_ix_sample_0_0,
    (SqwIXSource.serial_name, SqwIXSource.version): _parse_ix_source_2_0,
    ("unique_references_container", 1.0): _parse_unique_references_container_1_0,
    ("unique_objects_container", 1.0): _parse_unique_objects_container_1_0,
}


def _read_pix_block(sqw_io: LowLevelSqw) -> npt.NDArray[np.float32]:
    n_rows = sqw_io.read_u32()
    n_pixels = sqw_io.read_u64()
    return sqw_io.read_array((n_rows, n_pixels), np.dtype("float32"))


def _read_dnd_block(
    sqw_io: LowLevelSqw,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.uint64]]:
    # like metadata.axes.n_bins_all_dims?
    n_dims = sqw_io.read_u32()  # u32 not u8 as normal
    shape = tuple(sqw_io.read_u32() for _ in range(n_dims))
    values = sqw_io.read_array(shape, np.dtype("float64"))
    errors = sqw_io.read_array(shape, np.dtype("float64"))
    counts = sqw_io.read_array(shape, np.dtype("uint64"))
    return values, errors, counts



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import os
from dataclasses import replace
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any

import numpy as np
import pytest
import scipp as sc
from dateutil.parser import parse as parse_datetime

from scippneutron.io.sqw import (
    EnergyMode,
    Sqw,
    SqwDndMetadata,
    SqwIXExperiment,
    SqwIXNullInstrument,
    SqwIXSample,
    SqwIXSource,
    SqwLineAxes,
    SqwLineProj,
)

pace_neutrons = pytest.importorskip("pace_neutrons")


@pytest.fixture(scope="module")
def matlab() -> Any:
    try:
        return pace_neutrons.Matlab()
    except RuntimeError as e:
        if "No supported MATLAB" in e.args[0]:
            pytest.skip("MATLAB is unavailable")


@pytest.fixture
def dnd_metadata() -> SqwDndMetadata:
    return SqwDndMetadata(
        axes=SqwLineAxes(
            title="My Axes",
            label=["u1", "u2", "u3", "u4"],
            img_scales=[
                sc.scalar(1.0, unit="1/angstrom"),
                sc.scalar(1.0, unit="1/angstrom"),
                sc.scalar(1.0, unit="1/angstrom"),
                sc.scalar(1.0, unit="meV"),
            ],
            img_range=[
                sc.array(dims=["range"], values=[0.0, 1.0], unit="1/angstrom"),
                sc.array(dims=["range"], values=[0.0, 1.0], unit="1/angstrom"),
                sc.array(dims=["range"], values=[0.0, 1.0], unit="1/angstrom"),
                sc.array(dims=["range"], values=[0.0, 1.0], unit="meV"),
            ],
            # must be > 1
            n_bins_all_dims=sc.array(dims=["axis"], values=[2, 2, 2, 2], unit=None),
            single_bin_defines_iax=sc.array(dims=["axis"], values=[True] * 4),
            dax=sc.arange("axis", 4, unit=None),
            offset=[
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="meV"),
            ],
            changes_aspect_ratio=True,
            filename="dnd_axes",
            filepath="/dnd",
        ),
        proj=SqwLineProj(
            title="My Projection",
            lattice_spacing=sc.vector([2.86, 2.86, 2.86], unit="angstrom"),
            lattice_angle=sc.vector([90.0, 90.0, 90.0], unit="deg"),
            offset=[
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="1/angstrom"),
                sc.scalar(0.0, unit="meV"),
            ],
            label=["u1", "u2", "u3", "u4"],
            u=sc.vector([1.0, 0.0, 0.0], unit="1/angstrom"),
            v=sc.vector([0.0, 1.0, 0.0], unit="1/angstrom"),
            w=None,
            non_orthogonal=False,
            type="aaa",
        ),
    )


@pytest.fixture
def null_instrument() -> SqwIXNullInstrument:
    return SqwIXNullInstrument(
        name="Custom Instrument",
        source=SqwIXSource(
            name="My Source",
            target_name="The target",
            frequency=sc.scalar(13.4, unit="MHz"),
        ),
    )


@pytest.fixture
def sample() -> SqwIXSample:
    return SqwIXSample(
        name="Vibranium",
        lattice_spacing=sc.vector([2.86, 2.86, 2.86], unit="angstrom"),
        lattice_angle=sc.vector([90.0, 90.0, 90.0], unit="deg"),
    )


@pytest.fixture
def experiment_template() -> SqwIXExperiment:
    return SqwIXExperiment(
        run_id=0,
        efix=sc.scalar(1.2, unit="meV"),
        emode=EnergyMode.direct,
        en=sc.array(dims=["energy_transfer"], values=[-0.1, 0.3, 0.5], unit="meV"),
        psi=sc.scalar(0.4, unit="rad"),
        u=sc.vector([1.0, 0.0, 0.0], unit="1/angstrom"),
        v=sc.vector([0.0, 1.0, 0.0], unit="1/angstrom"),
        omega=sc.scalar(-0.01, unit="rad"),
        dpsi=sc.scalar(0.0, unit="rad"),
        gl=sc.scalar(1.2, unit="rad"),
        gs=sc.scalar(0.6, unit="rad"),
        filename="experiment1.nxspe",
        filepath="/data",
    )


@pytest.fixture
def pixel_data() -> sc.DataArray:
    rng = np.random.default_rng(9293)
    n_pixels = 76
    # Chosen numbers can all be represented in float32 to allow exact comparisons.
    return sc.DataArray(
        sc.array(
            dims=['obs'],
            values=rng.uniform(0, 100, n_pixels).astype('float32'),
            variances=rng.uniform(0.1, 1, n_pixels).astype('float32'),
            unit='count',
        ),
        coords={
            'idet': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(3),
            'irun': sc.arange('obs', 0, n_pixels, unit=None).astype(int) // sc.index(2),
            'ien': sc.arange('obs', 0, 2 * n_pixels, 2, unit=None).astype(int)
            // sc.index(10),
            'u1': sc.arange('obs', 0.0, n_pixels + 0.0, unit='1/Å'),
            'u2': sc.arange('obs', 1.0, n_pixels + 1.0, unit='1/Å'),
            'u3': sc.arange('obs', 2.0, n_pixels + 2.0, unit='1/Å'),
            'u4': sc.arange('obs', n_pixels, unit='meV') * 2,
        },
    )


def test_horace_roundtrip_main_header(
    matlab: Any,
    dnd_metadata: SqwDndMetadata,
    null_instrument: SqwIXNullInstrument,
    sample: SqwIXSample,
    experiment_template: SqwIXExperiment,
    pixel_data: sc.DataArray,
    tmp_path: Path,
) -> None:
    path = tmp_path / "roundtrip_main_header.sqw"
    (
        Sqw.build(path, title="Minimal test file")
        .add_default_instrument(null_instrument)
        .add_default_sample(sample)
        .add_empty_dnd_data(dnd_metadata)
        .add_pixel_data(pixel_data, experiments=[experiment_template])
        .create()
    )

    loaded_file = matlab.read_horace(os.fspath(path))
    main_header = loaded_file.main_header
    assert main_header.filename == path.name
    assert main_header.title == "Minimal test file"
    assert main_header.nfiles == 1
    assert (
        parse_datetime(main_header.creation_date).astimezone(tz=timezone.utc)
        - datetime.now(tz=timezone.utc)
    ) < timedelta(seconds=5)


def test_horace_roundtrip_null_instruments(
    matlab: Any,
    dnd_metadata: SqwDndMetadata,
    null_instrument: SqwIXNullInstrument,
    sample: SqwIXSample,
    experiment_template: SqwIXExperiment,
    pixel_data: sc.DataArray,
    tmp_path: Path,
) -> None:
    path = tmp_path / "roundtrip_null_instrument.sqw"
    (
        Sqw.build(path)
        .add_default_instrument(null_instrument)
        .add_default_sample(sample)
        .add_empty_dnd_data(dnd_metadata)
        .add_pixel_data(pixel_data, experiments=[experiment_template])
        .create()
    )

    loaded_file = matlab.read_horace(os.fspath(path))
    loaded_instruments = loaded_file.experiment_info.instruments.unique_objects
    assert np.array(loaded_instruments.n_objects).squeeze() == 1
    loaded = loaded_instruments[0]
    assert loaded.name == null_instrument.name
    assert loaded.source.name == null_instrument.source.name
    assert loaded.source.target_name == null_instrument.source.target_name
    assert loaded.source.frequency == null_instrument.source.frequency.value


def test_horace_roundtrip_sample(
    matlab: Any,
    dnd_metadata: SqwDndMetadata,
    null_instrument: SqwIXNullInstrument,
    sample: SqwIXSample,
    experiment_template: SqwIXExperiment,
    pixel_data: sc.DataArray,
    tmp_path: Path,
) -> None:
    path = tmp_path / "roundtrip_sample.sqw"
    (
        Sqw.build(path)
        .add_default_instrument(null_instrument)
        .add_default_sample(sample)
        .add_empty_dnd_data(dnd_metadata)
        .add_pixel_data(pixel_data, experiments=[experiment_template])
        .create()
    )

    loaded_file = matlab.read_horace(os.fspath(path))
    loaded_samples = loaded_file.experiment_info.samples.unique_objects
    assert loaded_samples.n_objects.squeeze() == 1
    loaded = loaded_samples[0]
    assert loaded.name == sample.name
    np.testing.assert_equal(
        loaded.alatt.squeeze(), sample.lattice_spacing.to(unit="angstrom").values
    )
    np.testing.assert_equal(
        loaded.angdeg.squeeze(), sample.lattice_angle.to(unit="deg").values
    )


def test_horace_roundtrip_experiment(
    matlab: Any,
    dnd_metadata: SqwDndMetadata,
    null_instrument: SqwIXNullInstrument,
    sample: SqwIXSample,
    experiment_template: SqwIXExperiment,
    pixel_data: sc.DataArray,
    tmp_path: Path,
) -> None:
    path = tmp_path / "roundtrip_experiment.sqw"
    (
        Sqw.build(path)
        .add_default_instrument(null_instrument)
        .add_default_sample(sample)
        .add_empty_dnd_data(dnd_metadata)
        .add_pixel_data(pixel_data, experiments=[experiment_template])
        .create()
    )

    loaded_file = matlab.read_horace(os.fspath(path))
    loaded_experiments = loaded_file.experiment_info.expdata
    assert matlab.numel(loaded_experiments).squeeze() == 1
    loaded = loaded_experiments[0]
    expected = experiment_template

    assert loaded.run_id.squeeze() == expected.run_id + 1
    np.testing.assert_equal(loaded.efix, expected.efix.to(unit="meV").value)
    assert loaded.emode == expected.emode.value
    np.testing.assert_equal(loaded.en.squeeze(), expected.en.to(unit="meV").values)
    np.testing.assert_equal(loaded.psi.squeeze(), expected.psi.to(unit="rad").value)
    np.testing.assert_equal(loaded.u.squeeze(), expected.u.to(unit="1/angstrom").values)
    np.testing.assert_equal(loaded.v.squeeze(), expected.v.to(unit="1/angstrom").values)
    np.testing.assert_equal(loaded.omega.squeeze(), expected.omega.to(unit="rad").value)
    np.testing.assert_equal(loaded.dpsi.squeeze(), expected.dpsi.to(unit="rad").value)
    np.testing.assert_equal(loaded.gl.squeeze(), expected.gl.to(unit="rad").value)
    np.testing.assert_equal(loaded.gs.squeeze(), expected.gs.to(unit="rad").value)
    assert loaded.filename == expected.filename
    assert loaded.filepath == expected.filepath


def test_horace_roundtrip_experiment_indirect(
    matlab: Any,
    dnd_metadata: SqwDndMetadata,
    null_instrument: SqwIXNullInstrument,
    sample: SqwIXSample,
    pixel_data: sc.DataArray,
    tmp_path: Path,
) -> None:
    experiment_template = SqwIXExperiment(
        run_id=0,
        efix=sc.array(dims=['detector'], values=[0.5, 0.6, 0.8, 0.9, 1.1], unit="meV"),
        emode=EnergyMode.direct,
        en=sc.array(
            dims=["detector", "energy_transfer"],
            values=[[-0.1, 0.3], [-0.2, 0.2], [0.0, 0.6], [0.1, 0.7], [0.3, 0.7]],
            unit="meV",
        ),
        psi=sc.scalar(0.4, unit="rad"),
        u=sc.vector([1.0, 0.0, 0.0], unit="1/angstrom"),
        v=sc.vector([0.0, 1.0, 0.0], unit="1/angstrom"),
        omega=sc.scalar(-0.01, unit="rad"),
        dpsi=sc.scalar(0.0, unit="rad"),
        gl=sc.scalar(1.2, unit="rad"),
        gs=sc.scalar(0.6, unit="rad"),
        filename="experiment1.nxspe",
        filepath="/data",
    )

    path = tmp_path / "roundtrip_experiment_indirect.sqw"
    (
        Sqw.build(path)
        .add_default_instrument(null_instrument)
        .add_default_sample(sample)
        .add_empty_dnd_data(dnd_metadata)
        .add_pixel_data(pixel_data, experiments=[experiment_template])
        .create()
    )

    loaded_file = matlab.read_horace(os.fspath(path))
    loaded_experiments = loaded_file.experiment_info.expdata
    assert matlab.numel(loaded_experiments).squeeze() == 1
    loaded = loaded_experiments[0]
    expected = experiment_template

    assert loaded.run_id.squeeze() == expected.run_id + 1
    np.testing.assert_equal(loaded.efix.squeeze(), expected.efix.to(unit="meV").values)
    assert loaded.emode == expected.emode.value
    np.testing.assert_equal(
        loaded.en.reshape(expected.en.shape), expected.en.to(unit="meV").values
    )
    np.testing.assert_equal(loaded.psi.squeeze(), expected.psi.to(unit="rad").value)
    np.testing.assert_equal(loaded.u.squeeze(), expected.u.to(unit="1/angstrom").values)
    np.testing.assert_equal(loaded.v.squeeze(), expected.v.to(unit="1/angstrom").values)
    np.testing.assert_equal(loaded.omega.squeeze(), expected.omega.to(unit="rad").value)
    np.testing.assert_equal(loaded.dpsi.squeeze(), expected.dpsi.to(unit="rad").value)
    np.testing.assert_equal(loaded.gl.squeeze(), expected.gl.to(unit="rad").value)
    np.testing.assert_equal(loaded.gs.squeeze(), expected.gs.to(unit="rad").value)
    assert loaded.filename == expected.filename
    assert loaded.filepath == expected.filepath


def test_horace_roundtrip_pixels(
    matlab: Any,
    dnd_metadata: SqwDndMetadata,
    null_instrument: SqwIXNullInstrument,
    sample: SqwIXSample,
    experiment_template: SqwIXExperiment,
    pixel_data: sc.DataArray,
    tmp_path: Path,
) -> None:
    path = tmp_path / "roundtrip_pixels.sqw"

    experiments = [
        replace(experiment_template, run_id=0, filename="experiment_1.nxspe"),
        replace(experiment_template, run_id=1, filename="experiment_2.nxspe"),
    ]
    rng = np.random.default_rng(526)
    pixel_data.coords['irun'] = sc.array(
        dims=['obs'], values=rng.integers(0, 2, len(pixel_data)), unit=None
    )

    (
        Sqw.build(path, title="Pixel test file")
        .add_default_instrument(null_instrument)
        .add_default_sample(sample)
        .add_empty_dnd_data(dnd_metadata)
        .add_pixel_data(pixel_data, experiments=experiments)
        .create()
    )

    loaded = matlab.read_horace(os.fspath(path))
    np.testing.assert_equal(loaded.pix.u1.squeeze(), pixel_data.coords['u1'].values)
    np.testing.assert_equal(loaded.pix.u2.squeeze(), pixel_data.coords['u2'].values)
    np.testing.assert_equal(loaded.pix.u3.squeeze(), pixel_data.coords['u3'].values)
    np.testing.assert_equal(loaded.pix.dE.squeeze(), pixel_data.coords['u4'].values)
    np.testing.assert_equal(
        loaded.pix.run_idx.squeeze(), pixel_data.coords['irun'].values
    )
    np.testing.assert_equal(
        loaded.pix.detector_idx.squeeze(), pixel_data.coords['idet'].values
    )
    np.testing.assert_equal(
        loaded.pix.energy_idx.squeeze(), pixel_data.coords['ien'].values
    )
    np.testing.assert_equal(loaded.pix.signal.squeeze(), pixel_data.values)
    np.testing.assert_equal(loaded.pix.variance.squeeze(), pixel_data.variances)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""Intermediate representation for SQW objects."""

from __future__ import annotations

import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Any, ClassVar, TypeVar

import numpy.typing as npt

_T = TypeVar("_T")


class TypeTag(enum.Enum):
    """Single byte tag to identify types in SQW files."""

    # Gaps in values are unsupported types.
    logical = 0
    char = 1
    f64 = 3
    f32 = 4
    i8 = 5
    u8 = 6
    i32 = 9
    u32 = 10
    i64 = 11
    u64 = 12
    cell = 23
    struct = 24
    serializable = 32  # objects that 'serialize themselves'


@dataclass(kw_only=True)
class ObjectArray:
    shape: tuple[int, ...]
    data: list[Object] | npt.NDArray[Any]
    ty: TypeTag


@dataclass(kw_only=True)
class CellArray:
    shape: tuple[int, ...]
    # nested object array to encode types of each item
    data: list[ObjectArray | CellArray]
    ty: ClassVar[TypeTag] = TypeTag.cell


@dataclass(kw_only=True)
class Struct:
    field_names: tuple[str, ...]
    field_values: CellArray
    ty: ClassVar[TypeTag] = TypeTag.struct

    def to_object_array(self) -> ObjectArray:
        return ObjectArray(
            ty=self.ty,
            shape=(1,),
            data=[self],
        )


@dataclass()
class String:
    value: str
    ty: ClassVar[TypeTag] = TypeTag.char


@dataclass()
class F64:
    value: float
    ty: ClassVar[TypeTag] = TypeTag.f64


@dataclass()
class U64:
    value: int
    ty: ClassVar[TypeTag] = TypeTag.u64


@dataclass()
class U32:
    value: int
    ty: ClassVar[TypeTag] = TypeTag.u32


@dataclass()
class U8:
    value: int
    ty: ClassVar[TypeTag] = TypeTag.u8


@dataclass()
class Logical:
    value: bool
    ty: ClassVar[TypeTag] = TypeTag.logical


# Not a dedicated type in SQW. Used here to encode NumPy arrays where list[ir.Object]
# is not efficient enough.
@dataclass()
class Array:
    value: npt.NDArray[Any]
    ty: TypeTag


# Not supported by SQW but represented here to simplify serialization.
@dataclass()
class Datetime:
    value: datetime
    ty: ClassVar[TypeTag] = TypeTag.char


Object = Struct | String | F64 | U64 | U32 | U8 | Logical | Array | Datetime


class Serializable(ABC):
    @abstractmethod
    def _serialize_to_dict(self) -> dict[str, Object | ObjectArray | CellArray]: ...

    def serialize_to_ir(self) -> Struct:
        fields = self._serialize_to_dict()
        return Struct(
            field_names=tuple(fields),
            field_values=CellArray(
                shape=(len(fields), 1),  # HORACE uses a 2D array
                data=[_serialize_field(field) for field in fields.values()],
            ),
        )

    def prepare_for_serialization(self: _T, filename: str, filepath: str) -> _T:  # noqa: PYI019
        return self


def _serialize_field(
    field: Object | ObjectArray | CellArray,
) -> ObjectArray | CellArray:
    if isinstance(field, ObjectArray | CellArray):
        return field
    if isinstance(field, Datetime):
        field = String(value=field.value.isoformat(timespec="seconds"))
    if isinstance(field, String):
        # TODO do we need to set the shape to empty?
        #  do we need to treat missing strings differently from empty strings?
        return ObjectArray(
            ty=field.ty, shape=(len(field.value),) if field.value else (), data=[field]
        )
    if isinstance(field, Array):
        return ObjectArray(ty=field.ty, shape=field.value.shape[::-1], data=field.value)
    return ObjectArray(ty=field.ty, shape=(1,), data=[field])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""Implementations of readers and writers for SQW object types."""

from collections.abc import Callable, Sequence
from typing import Any, Generic, TypeVar

import numpy as np
import numpy.typing as npt

from . import _ir as ir
from ._low_level_io import LowLevelSqw

_Shape = tuple[int, ...]
_T = TypeVar("_T")
_AnyObjectList = (
    Sequence[ir.Object] | list[ir.ObjectArray | ir.CellArray] | npt.NDArray[Any]
)
_ObjectReader = Callable[[LowLevelSqw, _Shape], _AnyObjectList]
_ObjectWriter = Callable[[LowLevelSqw, _AnyObjectList], None]


class _IORegistry(Generic[_T]):
    def __init__(self, action: str) -> None:
        self._action = action
        self._registry: dict[ir.TypeTag, _T] = {}

    def add(self, ty: ir.TypeTag) -> Callable[[_T], _T]:
        def add_impl(impl: _T) -> _T:
            if ty in self._registry:
                raise ValueError(f"Duplicate registration for {self._action} type {ty}")
            self._registry[ty] = impl
            return impl

        return add_impl

    def get(self, ty: ir.TypeTag, pos: int) -> _T:
        try:
            return self._registry[ty]
        except KeyError:
            raise ValueError(
                f"No {self._action} for SQW type {ty} as position {pos}"
            ) from None


_READERS = _IORegistry[_ObjectReader]("reader")
_WRITERS = _IORegistry[_ObjectWriter]("writer")


def read_object_array(sqw_io: LowLevelSqw) -> ir.ObjectArray | ir.CellArray:
    position = sqw_io.position
    ty = ir.TypeTag(sqw_io.read_u8())
    if ty == ir.TypeTag.serializable:  # TODO
        # raise RuntimeError(f'!!!! {ty.value} {sqw_io.position-1}')
        # This type object does not encode a shape, so just attempt
        # to read its contents.
        return read_object_array(sqw_io)

    shape = _read_shape(sqw_io)
    reader = _READERS.get(ty, position)
    data = reader(sqw_io, shape)
    if ty == ir.TypeTag.cell:
        return ir.CellArray(shape=shape, data=data)  # type: ignore[arg-type]
    return ir.ObjectArray(ty=ty, shape=shape, data=data)  # type: ignore[arg-type]


def write_object_array(
    sqw_io: LowLevelSqw, objects: ir.ObjectArray | ir.CellArray
) -> None:
    position = sqw_io.position

    if objects.ty == ir.TypeTag.struct:
        structs = objects.data
        if len(structs) == 1:
            from ._sqw import AbortParse, _get_scalar_struct_field

            try:
                # TODO use better mechanism
                name = _get_scalar_struct_field(structs[0], "serial_name")
                if name.startswith("IX_"):
                    sqw_io.write_u8(32)
            except AbortParse:
                pass

    sqw_io.write_u8(objects.ty.value)
    sqw_io.write_u8(len(objects.shape))  # TODO correct for list of structs?
    for size in objects.shape:
        sqw_io.write_u32(size)

    writer = _WRITERS.get(objects.ty, position)
    writer(sqw_io, objects.data)


@_READERS.add(ir.TypeTag.char)
def _read_char_arrays(sqw_io: LowLevelSqw, shape: _Shape) -> list[ir.Object]:
    # TODO is the str length shape[0] or shape[-1]?
    if not shape:
        return [ir.String("")]
    return [ir.String(sqw_io.read_n_chars(shape[0])) for _ in range(_volume(shape[1:]))]


@_WRITERS.add(ir.TypeTag.char)
def _write_char_array(sqw_io: LowLevelSqw, objects: _AnyObjectList) -> None:
    for obj in objects:
        chars: ir.String = obj  # type: ignore[assignment]
        sqw_io.write_chars(chars.value)


@_READERS.add(ir.TypeTag.cell)
def _read_cell(sqw_io: LowLevelSqw, shape: _Shape) -> Any:
    return [read_object_array(sqw_io) for _ in range(_volume(shape))]


@_WRITERS.add(ir.TypeTag.cell)
def _write_cell(sqw_io: LowLevelSqw, objects: _AnyObjectList) -> None:
    obj_arrays: list[ir.ObjectArray] = objects  # type: ignore[assignment]
    for obj in obj_arrays:
        write_object_array(sqw_io, obj)


# Arrays of struct are encoded with both the shape of the object array and the shape of
# the child cell array. Note the check of the shape.
@_READERS.add(ir.TypeTag.struct)
def _read_struct(sqw_io: LowLevelSqw, shape: _Shape) -> Sequence[ir.Object]:
    position = sqw_io.position
    if not shape:
        return []
    struct = _read_single_struct(sqw_io)
    if shape == (1,):
        expected_shape = (len(struct.field_names), *shape)
    else:  # There is an extra 1 in the shape.
        expected_shape = (len(struct.field_names), 1, *shape)
    if struct.field_values.shape != expected_shape:
        raise RuntimeError(
            f"While reading an array of structs with shape {shape} at position "
            f"{position}, expected the cell array of field values to have shape "
            f"{expected_shape}, but got {struct.field_values.shape}."
        )

    structs = []
    n_fields = len(struct.field_names)
    for i in range(_volume(shape)):
        fields = struct.field_values.data[i * n_fields : (i + 1) * n_fields]
        structs.append(
            ir.Struct(
                field_names=struct.field_names,
                field_values=ir.CellArray(
                    shape=(n_fields, 1),
                    data=fields,
                ),
            )
        )

    return structs


def _read_single_struct(sqw_io: LowLevelSqw) -> ir.Struct:
    # Read a single object tagged as a struct.
    # This can include an n-dimensional cell array for the struct values.
    n_fields = sqw_io.read_u32()
    field_name_sizes = [sqw_io.read_u32() for _ in range(n_fields)]
    field_names = tuple(sqw_io.read_n_chars(size) for size in field_name_sizes)
    field_values: ir.CellArray = _expect_ty(ir.TypeTag.cell, read_object_array(sqw_io))  # type: ignore[assignment]
    return ir.Struct(field_names=field_names, field_values=field_values)


@_WRITERS.add(ir.TypeTag.struct)
def _write_struct(sqw_io: LowLevelSqw, objects: _AnyObjectList) -> None:
    if not objects:
        return
    structs: list[ir.Struct] = objects  # type: ignore[assignment]
    n_structs = len(structs)
    n_fields = len(structs[0].field_names)
    cell_array_shape = (n_fields, 1) if n_structs == 1 else (n_fields, 1, n_structs)

    fields = []
    for struct in structs:
        fields.extend(struct.field_values.data)

    combined = ir.Struct(
        field_names=structs[0].field_names,
        field_values=ir.CellArray(shape=cell_array_shape, data=fields),
    )
    _write_single_struct(sqw_io, combined)


def _write_single_struct(sqw_io: LowLevelSqw, struct: ir.Struct) -> None:
    sqw_io.write_u32(len(struct.field_names))
    for name in struct.field_names:
        sqw_io.write_u32(len(name))
    for name in struct.field_names:
        sqw_io.write_chars(name)
    write_object_array(sqw_io, struct.field_values)


@_READERS.add(ir.TypeTag.f64)
def _read_f64(
    sqw_io: LowLevelSqw, shape: _Shape
) -> list[ir.Object] | npt.NDArray[np.float64]:
    data = sqw_io.read_array(shape, np.dtype("float64"))
    if data.size == 1:
        return [ir.F64(data.squeeze().item())]
    return data


@_WRITERS.add(ir.TypeTag.f64)
def _write_f64(sqw_io: LowLevelSqw, objects: _AnyObjectList) -> None:
    if isinstance(objects, np.ndarray):
        sqw_io.write_array(objects)
    else:
        for obj in objects:
            f64: ir.F64 = obj  # type: ignore[assignment]
            sqw_io.write_f64(f64.value)


@_READERS.add(ir.TypeTag.logical)
def _read_logical(sqw_io: LowLevelSqw, shape: _Shape) -> list[ir.Object]:
    return [ir.Logical(sqw_io.read_logical()) for _ in range(_volume(shape))]


@_WRITERS.add(ir.TypeTag.logical)
def _write_logical(sqw_io: LowLevelSqw, objects: _AnyObjectList) -> None:
    for obj in objects:
        logical: ir.Logical = obj  # type: ignore[assignment]
        sqw_io.write_logical(logical.value)


def _read_shape(sqw_io: LowLevelSqw) -> _Shape:
    n_dims = sqw_io.read_u8()
    return tuple(sqw_io.read_u32() for _ in range(n_dims))


_O = TypeVar("_O", bound=ir.Object | ir.ObjectArray | ir.CellArray)


def _expect_ty(ty: ir.TypeTag, obj: _O) -> _O:
    if obj.ty != ty:
        raise TypeError(f"Expected {ty}, got {obj.ty}")
    return obj


def _volume(shape: _Shape) -> int:
    return int(np.prod(shape))
